"""This module allows to estimate the feature relevance of inputs with respect to a given model. While these models can
be blackbox prediction models, it is also possible to explain causal mechanisms with respect to the direct parents.
In these cases, it would be possible to incorporate the noise to represent the part of the generation process that
cannot be explained by the parents.

Functions in this module should be considered experimental, meaning there might be breaking API changes in the future.
"""
from typing import Any, Callable, Dict, Optional, Tuple, Union

import numpy as np
import pandas as pd

from dowhy.gcm.cms import StructuralCausalModel
from dowhy.gcm.fcms import ProbabilityEstimatorModel
from dowhy.gcm.fitting_sampling import draw_samples
from dowhy.gcm.graph import get_ordered_predecessors, is_root_node, validate_node
from dowhy.gcm.shapley import ShapleyConfig, estimate_shapley_values
from dowhy.gcm.stats import marginal_expectation
from dowhy.gcm.util.general import shape_into_2d, variance_of_deviations, variance_of_matching_values


def parent_relevance(
    causal_model: StructuralCausalModel,
    target_node: Any,
    parent_samples: Optional[pd.DataFrame] = None,
    subset_scoring_func: Optional[Callable[[np.ndarray, np.ndarray], Union[np.ndarray, float]]] = None,
    num_samples_randomization: int = 5000,
    num_samples_baseline: int = 500,
    max_batch_size: int = 100,
    shapley_config: Optional[ShapleyConfig] = None,
) -> Tuple[Dict[Any, Any], np.ndarray]:
    """Estimates the distribution based relevance of the direct parents of the given target_node. This is, the
    relevance of direct parents as input features of the the underlying causal model of target_node. Here, the
    unobserved noise is considered as a direct parent (input) as well. Samples utilized for the estimation are drawn
    from the given causal graph.

    By default, the used subset_scoring_func is based on the variance between Y and Y', where Y is the outputs of
    the causal model and Y' the outputs of the models when certain features are randomized. In case of continuous data,
    the feature relevance adds up to Var(Y - Y').

    Note: The feature relevance based on the distribution cannot be directly compared with the feature relevance for
    single samples. If this is desired, the set function needs to be defined accordingly.

    Related paper:
    Janzing, D., Minorics, L., & Bloebaum, P. (2020).
    Feature relevance quantification in explainable AI: A causal problem.
    In International Conference on Artificial Intelligence and Statistics (pp. 2907-2916). PMLR.

    :param causal_model: The fitted structural causal model.
    :param target_node: Node with the causal model of interest.
    :param parent_samples: Samples for the parents of the given target_node. If None is given, new samples are
                           generated based on the graph. These samples are used for randomizing features that are not in the subset.
    :param subset_scoring_func: Set function for estimating the quantity of interest based on the model outcomes. This function
                                expects two inputs; the outcome of the causal model for some samples if certain features are permuted and the
                                outcome of the model for the same samples when no features were permuted. The set functions represents the
                                comparison between the samples, for instance, the variance of deviations. This is then used as the 'characteristic function'
                                in coalition games when estimating the Shapley values.
    :param num_samples_randomization: Number of samples used as background parent samples for evaluating the set function.
                                      If no parent_samples are given, this represents the number of generated samples from
                                      the joint distribution of the parents and are used for randomizing features that are
                                      not in the subset. Consider increasing this number for more accurate results or reducing it for less memory consumption and faster runtime.
    :param num_samples_baseline: Number of samples on which the set functions are evaluated on. These samples are used as fixed observations for
                                 parents that are in the subset. Consider increasing this number for more accurate results or reducing it for less memory consumption and faster runtime.
    :param max_batch_size: Maximum batch size for estimating multiple predictions at once. This has a significant influence on the
                          overall memory usage. If set to -1, all samples are used in one batch.
    :param shapley_config: :class:`~dowhy.gcm.shapley.ShapleyConfig` for the Shapley estimator.
    :return: There are two return vales. A dictionary with the feature relevance for each direct parent of the given
             target_node and the feature relevance of noise.
    """
    validate_node(causal_model.graph, target_node)

    if is_root_node(causal_model.graph, target_node):
        raise ValueError(
            "Cannot compute feature relevance of parents for the target node %s as it is a root node."
            "It does not have parents." % target_node
        )

    ordered_predecessors = get_ordered_predecessors(causal_model.graph, target_node)

    if parent_samples is None:
        parent_samples = draw_samples(causal_model, max(num_samples_randomization, num_samples_baseline))[
            ordered_predecessors
        ]

    if subset_scoring_func is None:
        if isinstance(causal_model.causal_mechanism(target_node), ProbabilityEstimatorModel):
            subset_scoring_func = variance_of_matching_values
        else:
            subset_scoring_func = variance_of_deviations

    parent_samples = shape_into_2d(parent_samples[ordered_predecessors].to_numpy())
    noise_samples = shape_into_2d(
        causal_model.causal_mechanism(target_node).draw_noise_samples(parent_samples.shape[0])
    )
    samples_features = np.column_stack([parent_samples.astype(object), noise_samples.astype(object)])

    def model(X: np.ndarray) -> np.ndarray:
        return causal_model.causal_mechanism(target_node).evaluate(
            X[:, : -noise_samples.shape[1]], X[:, -noise_samples.shape[1] :]
        )

    shapley_vales = feature_relevance_distribution(
        model,
        feature_samples=samples_features,
        subset_scoring_func=subset_scoring_func,
        max_num_samples_randomization=num_samples_randomization,
        max_num_baseline_samples=num_samples_baseline,
        max_batch_size=max_batch_size,
        shapley_config=shapley_config,
    )

    result = {(predecessor, target_node): shapley_vales[i] for i, predecessor in enumerate(ordered_predecessors)}

    return result, shapley_vales[-noise_samples.shape[1] :]


def feature_relevance_distribution(
    prediction_method: Callable[[np.ndarray], np.ndarray],
    feature_samples: np.ndarray,
    subset_scoring_func: Callable[[np.ndarray, np.ndarray], Union[np.ndarray, float]],
    max_num_samples_randomization: int = 5000,
    max_num_baseline_samples: int = 500,
    max_batch_size: int = 100,
    randomize_features_jointly: bool = True,
    shapley_config: Optional[ShapleyConfig] = None,
) -> np.ndarray:
    """Estimates the population based feature relevance of the input features for the given prediction_method. This
    method uses all samples given in feature_samples by comparing the output of the prediction_method given certain
    features are randomized with the outputs when no features are randomized. The subset_scoring_func defines
    how these predictions are compared. For instance, the variance of deviations.

    If the randomized predictions should rather be compared to the original data, this has (and can) be defined via
    the set function by ignoring the second input parameter (the predicted values using all feauters). Instead, the
    original data can be used.

    Note: The distribution level relevance is estimated by taking the expectation of the outcome of the set functions
    when applied to multiple samples. Due to the linearity of the Shapley value estimation, this is equivalent to taking
    the expectation over the Shapley values.

    Related paper:
    Janzing, D., Minorics, L., & Bloebaum, P. (2020).
    Feature relevance quantification in explainable AI: A causal problem.
    In International Conference on Artificial Intelligence and Statistics (pp. 2907-2916). PMLR.

    :param prediction_method: A callable that is expected to return a prediction for given samples.
    :param feature_samples: Samples from the joint distribution.
    :param subset_scoring_func: Set function for estimating the quantity of interest based on the model outcomes. This function
                                expects two inputs; the outcome of the prediction model for some samples if certain features are permuted and the
                                outcome of the model for the same samples when no features were permuted. The set functions represents the
                                comparison between the samples, for instance, the variance of deviations. This is then used as the 'characteristic function'
                                in coalition games when estimating the Shapley values.
    :param max_num_samples_randomization: Maximum number of samples used for randomizing the feature that are not in the susbet. Consider increasing this
                                          number for more accurate results (if enough samples are available) or reducing it for less memory consumption and
                                          faster runtime.
    :param max_num_baseline_samples: Maximum number of samples on which the set function is evaluated on. These samples are used as fixed observations for
                                     features that are in the subset. For instance, in case of taking the mean as set_function_summary_func, this defines the maximum number
                                     of samples used to estimate the mean. Consider increasing this number for more accurate results (if enough samples are
                                     available) or reducing it for less memory consumption and faster runtime.
    :param max_batch_size: Maximum batch size for a estimating the predictions. This has a significant influence on the
                           overall memory usage. If set to -1, all samples are used in one batch.
    :param randomize_features_jointly: If set to True, features that are not in a subset are jointly permuted.
                                       Note that this still represents an interventional distribution. If set to False, features that are not in a subset
                                       are independently permuted. Note: The theory in the linked publication assumes that this is set to True.
    :param shapley_config: Config for the Shapley estimator.
    :return: A numpy array with the feature relevance of each input feature.
    """
    feature_samples = shape_into_2d(feature_samples)

    if shapley_config is None:
        shapley_config = ShapleyConfig()

    baseline_samples = feature_samples[
        np.random.choice(
            feature_samples.shape[0], min(max_num_baseline_samples, feature_samples.shape[0]), replace=False
        )
    ]
    feature_samples = feature_samples[
        np.random.choice(
            feature_samples.shape[0], min(max_num_samples_randomization, feature_samples.shape[0]), replace=False
        )
    ]

    return feature_relevance_sample(
        prediction_method,
        feature_samples,
        baseline_samples,
        subset_scoring_func,
        None,
        True,
        max_batch_size,
        randomize_features_jointly,
        shapley_config,
    )


def feature_relevance_sample(
    prediction_method: Callable[[np.ndarray], np.ndarray],
    feature_samples: np.ndarray,
    baseline_samples: np.ndarray,
    subset_scoring_func: Callable[[np.ndarray, np.ndarray], Union[np.ndarray, float]],
    baseline_target_values: Optional[np.ndarray] = None,
    average_set_function: bool = False,
    max_batch_size: int = 100,
    randomize_features_jointly: bool = True,
    shapley_config: Optional[ShapleyConfig] = None,
) -> np.ndarray:
    """Estimates the feature relevance of the prediction_method for each sample in baseline_noise_samples. This
    method uses all samples given in feature_samples as 'background' samples. This is, they should represent samples
    from the joint distribution of the input features. The subset_scoring_func defines the comparison between the
    output of the prediction_method when certain features are randomized and the outputs when no features are
    randomized. The most common function would be the difference between the expectations.

    If the randomized predictions should rather be compared to the original data, this has (and can) be defined via
    the set function by ignoring the second input parameter (the predicted values using all feauters). Instead, the
    original data can be used.

    Related paper:
    Janzing, D., Minorics, L., & Bloebaum, P. (2020).
    Feature relevance quantification in explainable AI: A causal problem.
    In International Conference on Artificial Intelligence and Statistics (pp. 2907-2916). PMLR.

    :param prediction_method: A callable that is expected to return a prediction for given samples.
    :param feature_samples: Samples from the joint distribution. These are used as 'background samples' to randomize features that are not in a subset.
    :param baseline_samples: Samples for which the feature relevance should be estimated.
    :param subset_scoring_func: Set function for estimating the quantity of interest based on the model outcomes. This function
                                expects two inputs; the outcome of the prediction model for some samples if certain features are permuted and the
                                outcome of the model for the same samples when no features were permuted. A typical choice for regression models
                                would be the difference between expectations. This is then used as the 'characteristic function'
                                in coalition games when estimating the Shapley values.
    :param baseline_target_values: These baseline values are compared with the subset specific outcomes of the prediction method. If set to None (default),
                                   the baseline values are the outcomes of the given prediction_method applied to the baseline_noise_samples, i.e. the outcome of the empty subset.
    :param max_batch_size: Maximum batch size for a estimating the predictions. This has a significant influence on the
                           overall memory usage. If set to -1, all samples are used in one batch.
    :param average_set_function: If set to True, the averaged result of the set function applied to each sample of
                                 interest is used for estimating the Shapley values. If set to False, Shapley values for each sample of interest
                                 are estimated separately.
    :param randomize_features_jointly: If set to True, features that are not in a subset are jointly permuted.
                                       Note that this still represents an interventional distribution. If set to False, features that are not in a subset
                                       are independently permuted. Note: The theory in the linked publication assumes that this is set to True.
    :param shapley_config: Config for the Shapley estimator.
    :return: A numpy array with the feature relevance for each sample in baseline_noise_samples.
    """
    feature_samples = shape_into_2d(feature_samples)
    if baseline_target_values is None:
        baseline_target_values = prediction_method(baseline_samples)
    else:
        if baseline_samples.shape[0] != baseline_target_values.shape[0]:
            raise ValueError(
                "Samples of interest and the given baseline values need to have the same sample size! "
                "Make sure that the given baseline values correspond to the samples of interest."
            )

    if shapley_config is None:
        shapley_config = ShapleyConfig()

    def single_sample_set_function(subset: np.ndarray) -> Union[np.ndarray, float]:
        results = np.zeros(baseline_target_values.shape[0])
        predictions = marginal_expectation(
            prediction_method,
            feature_samples=feature_samples,
            baseline_samples=baseline_samples,
            baseline_feature_indices=np.arange(0, feature_samples.shape[1])[subset == 1],
            return_averaged_results=False,
            feature_perturbation="randomize_columns_jointly"
            if randomize_features_jointly
            else "randomize_columns_independently",
            max_batch_size=max_batch_size,
        )

        for i in range(baseline_target_values.shape[0]):
            results[i] = subset_scoring_func(predictions[i], baseline_target_values[i])

        if average_set_function:
            return np.mean(results)
        else:
            return results

    return estimate_shapley_values(single_sample_set_function, feature_samples.shape[1], shapley_config)



"""Functions in this module should be considered experimental, meaning there might be breaking API changes in the
future.
"""

from typing import Optional

import numpy as np
from sklearn.kernel_approximation import Nystroem
from sklearn.metrics import euclidean_distances

from dowhy.gcm.util.general import shape_into_2d


def apply_rbf_kernel(X: np.ndarray, precision: Optional[float] = None) -> np.ndarray:
    """
    Estimates the RBF (Gaussian) kernel for the given input data.

    :param X: Input data.
    :param precision: Specific precision matrix for the RBF kernel. If None is given, this is inferred from the data.
    :return: The outcome of applying a RBF (Gaussian) kernel on the data.
    """
    X = shape_into_2d(X)

    distance_matrix = euclidean_distances(X, squared=True)

    if precision is None:
        precision = _median_based_precision(distance_matrix)

    return np.exp(-precision * distance_matrix)


def apply_rbf_kernel_with_adaptive_precision(X: np.ndarray) -> np.ndarray:
    """Estimates the RBF (Gaussian) kernel for the given input data. Here, each column is scaled by an individual
    precision parameter which is automatically inferred from the data.

    :param X: Input data.
    :return: The outcome of applying a RBF (Gaussian) kernel on the data.
    """
    X = shape_into_2d(X)

    result = np.ones((X.shape[0], X.shape[0]))
    for i in range(X.shape[1]):
        distance_matrix = euclidean_distances(X, squared=True)
        result *= np.exp(-_median_based_precision(distance_matrix) * distance_matrix)

    return result


def apply_delta_kernel(X: np.ndarray) -> np.ndarray:
    """Applies the delta kernel, i.e. the distance is 1 if two entries are equal and 0 otherwise.

    :param X: Input data.
    :return: The outcome of the delta-kernel, a binary distance matrix.
    """
    X = shape_into_2d(X)
    return np.array(list(map(lambda value: value == X, X))).reshape(X.shape[0], X.shape[0]).astype(np.float)


def approximate_rbf_kernel_features(
    X: np.ndarray, num_random_components: int, precision: Optional[float] = None
) -> np.ndarray:
    """Applies the Nystroem method to create a NxD (D << N) approximated RBF kernel map using a subset of the data,
    where N is the number of samples in X and D the number of components.

    :param X: Input data.
    :param num_random_components: Number of components D for the approximated kernel map.
    :param precision: Specific precision matrix for the RBF kernel. If None is given, this is inferred from the data.
    :return: A NxD approximated RBF kernel map, where N is the number of samples in X and D the number of components.
    """
    X = shape_into_2d(X)

    if precision is None:
        precision = _median_based_precision(euclidean_distances(X, squared=True))

    return Nystroem(kernel="rbf", gamma=precision, n_components=num_random_components).fit_transform(X)


def approximate_delta_kernel_features(X: np.ndarray, num_random_components: int) -> np.ndarray:
    """Applies the Nystroem method to create a NxD (D << N) approximated delta kernel map using a subset of the data,
    where N is the number of samples in X and D the number of components. The delta kernel gives 1 if two entries are
    equal and 0 otherwise.

    :param X: Input data.
    :param num_random_components: Number of components D for the approximated kernel map.
    :return: A NxD approximated RBF kernel map, where N is the number of samples in X and D the number of components.
    """
    X = shape_into_2d(X)

    def delta_function(x, y) -> float:
        return float(x == y)

    for i, unique_element in enumerate(np.unique(X)):
        X[X == unique_element] = i

    result = Nystroem(kernel=delta_function, n_components=num_random_components).fit_transform(X.astype(int))
    result[result != 0] = 1

    return result


def _median_based_precision(distances: np.ndarray) -> float:
    tmp = np.sqrt(distances)
    tmp = tmp - np.tril(tmp, -1)
    tmp = tmp.reshape(-1, 1)

    return 1 / np.median(tmp[tmp > 0])



import pytest
from pytest import mark

from dowhy.causal_estimators.propensity_score_matching_estimator import PropensityScoreMatchingEstimator

from .base import TestEstimator


@mark.usefixtures("fixed_seed")
class TestPropensityScoreMatchingEstimator(object):
    @mark.parametrize(
        [
            "error_tolerance",
            "Estimator",
            "num_common_causes",
            "num_instruments",
            "num_effect_modifiers",
            "num_treatments",
            "treatment_is_binary",
            "outcome_is_binary",
        ],
        [
            (
                0.3,
                PropensityScoreMatchingEstimator,
                [1, 2],
                [0],
                [
                    0,
                ],
                [
                    1,
                ],
                [
                    True,
                ],
                [
                    False,
                ],
            ),
        ],
    )
    def test_average_treatment_effect(
        self,
        error_tolerance,
        Estimator,
        num_common_causes,
        num_instruments,
        num_effect_modifiers,
        num_treatments,
        treatment_is_binary,
        outcome_is_binary,
    ):
        estimator_tester = TestEstimator(error_tolerance, Estimator)
        estimator_tester.average_treatment_effect_testsuite(
            num_common_causes=num_common_causes,
            num_instruments=num_instruments,
            num_effect_modifiers=num_effect_modifiers,
            num_treatments=num_treatments,
            treatment_is_binary=treatment_is_binary,
            outcome_is_binary=outcome_is_binary,
            confidence_intervals=[
                False,
            ],
            test_significance=[
                False,
            ],
            method_params={"num_simulations": 10, "num_null_simulations": 10},
        )



"""This module contains implementations of different density estimators.

Classes and functions in this module should be considered experimental, meaning there might be breaking API changes in
the future.
"""

from typing import Optional

import numpy as np
from sklearn.mixture import BayesianGaussianMixture
from sklearn.neighbors import KernelDensity

from dowhy.gcm.density_estimator import DensityEstimator
from dowhy.gcm.util.general import shape_into_2d


class GaussianMixtureDensityEstimator(DensityEstimator):
    """Represents a density estimator based on a Gaussian mixture model. The estimator uses the sklearn
    BayesianGaussianMixture model internally.
    """

    def __init__(self, num_components: Optional[int] = None) -> None:
        self._gmm_model = None
        self._num_components = num_components

    def fit(self, X: np.ndarray) -> None:
        if self._num_components is None:
            self._num_components = int(np.ceil(np.sqrt(X.shape[0] / 2)))

        self._gmm_model = BayesianGaussianMixture(n_components=self._num_components, covariance_type="full").fit(
            shape_into_2d(X)
        )

    def density(self, X: np.ndarray) -> np.ndarray:
        if self._gmm_model is None:
            raise RuntimeError("%s has not been fitted!" % self.__class__.__name__)

        # Note, the output of score_samples are log values.
        return np.exp(self._gmm_model.score_samples(shape_into_2d(X)))


class KernelDensityEstimator1D(DensityEstimator):
    """Represents a kernel based density estimator. The estimator uses the sklearn KernelDensity class internally."""

    def __init__(self) -> None:
        self._kde_model = None

    def fit(self, X: np.ndarray) -> None:
        X = shape_into_2d(X)
        self._validate_data(X)

        bandwidth = np.std(X) * np.power(4 / 3 / X.shape[0], 1 / 5)

        self._kde_model = KernelDensity(kernel="gaussian", bandwidth=bandwidth).fit(X.reshape(-1, 1))

    def _validate_data(self, X: np.ndarray) -> None:
        if X.shape[1] > 1:
            raise RuntimeError("%s only supports one dimensional data!" % self.__class__.__name__)

    def density(self, X: np.ndarray) -> np.ndarray:
        if self._kde_model is None:
            raise RuntimeError("%s has not been fitted!" % self.__class__.__name__)

        X = shape_into_2d(X)
        self._validate_data(X)

        # Note, the output of score_samples are log values.
        return np.exp(self._kde_model.score_samples(X.reshape(-1, 1)))



import inspect
from importlib import import_module
from typing import Any, Callable, List, Optional, Protocol, Union
from warnings import warn

import numpy as np
import pandas as pd
from numpy.distutils.misc_util import is_sequence

from dowhy.causal_estimator import CausalEstimate, CausalEstimator
from dowhy.causal_identifier import IdentifiedEstimand
from dowhy.utils.api import parse_state


class _EconmlEstimator(Protocol):
    def fit(self, *args, **kwargs):
        ...

    def effect(self, *args, **kwargs):
        ...

    def effect_interval(self, *args, **kwargs):
        ...

    def effect_inference(self, *args, **kwargs):
        ...

    def shap_values(self, *args, **kwargs):
        ...


class Econml(CausalEstimator):
    """Wrapper class for estimators from the EconML library.

    Supports additional parameters as listed below. For init and fit
    parameters of each estimator, refer to the EconML docs.

    """

    def __init__(
        self,
        identified_estimand: IdentifiedEstimand,
        econml_estimator: Union[_EconmlEstimator, str],
        test_significance: bool = False,
        evaluate_effect_strength: bool = False,
        confidence_intervals: bool = False,
        num_null_simulations: int = CausalEstimator.DEFAULT_NUMBER_OF_SIMULATIONS_STAT_TEST,
        num_simulations: int = CausalEstimator.DEFAULT_NUMBER_OF_SIMULATIONS_CI,
        sample_size_fraction: int = CausalEstimator.DEFAULT_SAMPLE_SIZE_FRACTION,
        confidence_level: float = CausalEstimator.DEFAULT_CONFIDENCE_LEVEL,
        need_conditional_estimates: Union[bool, str] = "auto",
        num_quantiles_to_discretize_cont_cols: int = CausalEstimator.NUM_QUANTILES_TO_DISCRETIZE_CONT_COLS,
        **kwargs,
    ):
        """
        :param identified_estimand: probability expression
            representing the target identified estimand to estimate.
        :param econml_methodname: Fully qualified name of econml estimator
            class. For example, 'econml.dml.DML'
        :param test_significance: Binary flag or a string indicating whether to test significance and by which method. All estimators support test_significance="bootstrap" that estimates a p-value for the obtained estimate using the bootstrap method. Individual estimators can override this to support custom testing methods. The bootstrap method supports an optional parameter, num_null_simulations. If False, no testing is done. If True, significance of the estimate is tested using the custom method if available, otherwise by bootstrap.
        :param evaluate_effect_strength: (Experimental) whether to evaluate the strength of effect
        :param confidence_intervals: Binary flag or a string indicating whether the confidence intervals should be computed and which method should be used. All methods support estimation of confidence intervals using the bootstrap method by using the parameter confidence_intervals="bootstrap". The bootstrap method takes in two arguments (num_simulations and sample_size_fraction) that can be optionally specified in the params dictionary. Estimators may also override this to implement their own confidence interval method. If this parameter is False, no confidence intervals are computed. If True, confidence intervals are computed by the estimator's specific method if available, otherwise through bootstrap
        :param num_null_simulations: The number of simulations for testing the
            statistical significance of the estimator
        :param num_simulations: The number of simulations for finding the
            confidence interval (and/or standard error) for a estimate
        :param sample_size_fraction: The size of the sample for the bootstrap
            estimator
        :param confidence_level: The confidence level of the confidence
            interval estimate
        :param need_conditional_estimates: Boolean flag indicating whether
            conditional estimates should be computed. Defaults to True if
            there are effect modifiers in the graph
        :param num_quantiles_to_discretize_cont_cols: The number of quantiles
            into which a numeric effect modifier is split, to enable
            estimation of conditional treatment effect over it.
        :param kwargs: (optional) Additional estimator-specific parameters

        """
        super().__init__(
            identified_estimand=identified_estimand,
            test_significance=test_significance,
            evaluate_effect_strength=evaluate_effect_strength,
            confidence_intervals=confidence_intervals,
            num_null_simulations=num_null_simulations,
            num_simulations=num_simulations,
            sample_size_fraction=sample_size_fraction,
            confidence_level=confidence_level,
            need_conditional_estimates=need_conditional_estimates,
            num_quantiles_to_discretize_cont_cols=num_quantiles_to_discretize_cont_cols,
            econml_estimator=econml_estimator,
            **kwargs,
        )

        if isinstance(econml_estimator, str):
            warn(
                "Using a string to specify the value for econml_estimator is now deprecated, please provide an instance of a econml object",
                DeprecationWarning,
                stacklevel=2,
            )
            estimator_class = self._get_econml_class_object(econml_estimator)
            self.estimator = estimator_class(**kwargs["init_params"])
        else:
            self.estimator = econml_estimator

        self.logger.info("INFO: Using EconML Estimator")
        self.identifier_method = self._target_estimand.identifier_method

    def fit(
        self,
        data: pd.DataFrame,
        treatment_name: str,
        outcome_name: str,
        effect_modifier_names: Optional[List[str]] = None,
        **kwargs,
    ):
        """
        Fits the estimator with data for effect estimation
        :param data: data frame containing the data
        :param treatment: name of the treatment variable
        :param outcome: name of the outcome variable
        :param effect_modifiers: Variables on which to compute separate
                    effects, or return a heterogeneous effect function. Not all
                    methods support this currently.
        """
        self._set_data(data, treatment_name, outcome_name)
        self._set_effect_modifiers(effect_modifier_names)

        # Save parameters for later refutter fitting
        self._econml_fit_params = kwargs

        self._observed_common_causes_names = self._target_estimand.get_backdoor_variables().copy()

        # Enforcing this ordering is necessary to feed through the propensity values from dataset
        self._observed_common_causes_names = [
            c for c in self._observed_common_causes_names if "propensity" not in c
        ] + sorted([c for c in self._observed_common_causes_names if "propensity" in c])

        # For metalearners only--issue a warning if w contains variables not in x
        if self.estimator.__module__.endswith("metalearners"):
            effect_modifier_names = []
            if self._effect_modifier_names is not None:
                effect_modifier_names = self._effect_modifier_names.copy()
            w_diff_x = [w for w in self._observed_common_causes_names if w not in effect_modifier_names]
            if len(w_diff_x) > 0:
                self.logger.warn(
                    "Concatenating common_causes and effect_modifiers and providing a single list of variables to metalearner estimator method, "
                    + self.estimator.__class__.__name__
                    + ". EconML metalearners accept a single X argument."
                )
                effect_modifier_names.extend(w_diff_x)
                # Override the effect_modifiers set in CausalEstimator.__init__()
                # Also only update self._effect_modifiers, and create a copy of self._effect_modifier_names
                # the latter can be used by other estimator methods later
                self._effect_modifiers = self._data[effect_modifier_names]
                self._effect_modifiers = pd.get_dummies(self._effect_modifiers, drop_first=True)
                self._effect_modifier_names = effect_modifier_names
            self.logger.debug("Effect modifiers: " + ",".join(effect_modifier_names))
        if self._observed_common_causes_names:
            self._observed_common_causes = self._data[self._observed_common_causes_names]
            self._observed_common_causes = pd.get_dummies(self._observed_common_causes, drop_first=True)
        else:
            self._observed_common_causes = None
        self.logger.debug("Back-door variables used:" + ",".join(self._observed_common_causes_names))
        # Instrumental variables names, if present
        # choosing the instrumental variable to use
        if getattr(self, "iv_instrument_name", None) is None:
            self.estimating_instrument_names = self._target_estimand.instrumental_variables
        else:
            self.estimating_instrument_names = parse_state(self.iv_instrument_name)
        if self.estimating_instrument_names:
            self._estimating_instruments = self._data[self.estimating_instrument_names]
            self._estimating_instruments = pd.get_dummies(self._estimating_instruments, drop_first=True)
        else:
            self._estimating_instruments = None

        self.symbolic_estimator = self.construct_symbolic_estimator(self._target_estimand)
        self.logger.info(self.symbolic_estimator)

        X = None
        W = None  # common causes/ confounders
        Z = None  # Instruments
        Y = self._outcome
        T = self._treatment
        if self._effect_modifiers is not None and len(self._effect_modifiers) > 0:
            X = self._effect_modifiers
        if self._observed_common_causes_names:
            W = self._observed_common_causes
        if self.estimating_instrument_names:
            Z = self._estimating_instruments
        named_data_args = {"Y": Y, "T": T, "X": X, "W": W, "Z": Z}

        # Calling the econml estimator's fit method
        estimator_argspec = inspect.getfullargspec(inspect.unwrap(self.estimator.fit))
        # As of v0.9, econml has some kewyord only arguments
        estimator_named_args = estimator_argspec.args + estimator_argspec.kwonlyargs
        estimator_data_args = {
            arg: named_data_args[arg] for arg in named_data_args.keys() if arg in estimator_named_args
        }
        self.estimator.fit(**estimator_data_args, **kwargs)

        return self

    def _get_econml_class_object(self, module_method_name, *args, **kwargs):
        # from https://www.bnmetrics.com/blog/factory-pattern-in-python3-simple-version
        try:
            (module_name, _, class_name) = module_method_name.rpartition(".")
            estimator_module = import_module(module_name)
            estimator_class = getattr(estimator_module, class_name)

        except (AttributeError, AssertionError, ImportError):
            raise ImportError(
                "Error loading {}.{}. Double-check the method name and ensure that all econml dependencies are installed.".format(
                    module_name, class_name
                )
            )
        return estimator_class

    def estimate_effect(
        self, data: pd.DataFrame = None, treatment_value: Any = 1, control_value: Any = 0, target_units=None, **_
    ):
        """
        data: dataframe containing the data on which treatment effect is to be estimated.
        treatment_value: value of the treatment variable for which the effect is to be estimated.
        control_value: value of the treatment variable that denotes its absence (usually 0)
        target_units: The units for which the treatment effect should be estimated.
                     It can be a DataFrame that contains values of the effect_modifiers and effect will be estimated only for this new data.
                     It can also be a lambda function that can be used as an index for the data (pandas DataFrame) to select the required rows.
        """
        if data is None:
            data = self._data
        self._target_units = target_units
        self._treatment_value = treatment_value
        self._control_value = control_value

        n_samples = self._treatment.shape[0]
        X = None  # Effect modifiers
        if self._effect_modifiers is not None and len(self._effect_modifiers) > 0:
            X = self._effect_modifiers

        X_test = X
        if X is not None:
            if type(target_units) is pd.DataFrame:
                X_test = target_units
            elif callable(target_units):
                filtered_rows = data.where(target_units)
                boolean_criterion = np.array(filtered_rows.notnull().iloc[:, 0])
                X_test = X[boolean_criterion]
        # Changing shape to a list for a singleton value
        self._treatment_value = parse_state(self._treatment_value)

        est = self.effect(X_test)
        ate = np.mean(est, axis=0)  # one value per treatment value

        if len(ate) == 1:
            ate = ate[0]

        if self._confidence_intervals:
            self.effect_intervals = self.effect_interval(X_test)
        else:
            self.effect_intervals = None

        estimate = CausalEstimate(
            estimate=ate,
            control_value=control_value,
            treatment_value=treatment_value,
            target_estimand=self._target_estimand,
            realized_estimand_expr=self.symbolic_estimator,
            cate_estimates=est,
            effect_intervals=self.effect_intervals,
            _estimator_object=self.estimator,
        )

        estimate.add_estimator(self)
        return estimate

    def _estimate_confidence_intervals(self, confidence_level=None, method=None):
        """Returns None if the confidence interval has not been calculated."""
        return self.effect_intervals

    def _do(self, x):
        raise NotImplementedError

    def construct_symbolic_estimator(self, estimand):
        expr = "b: " + ", ".join(estimand.outcome_variable) + "~"
        # TODO -- fix: we are actually conditioning on positive treatment (d=1)
        if self.estimator.__module__.endswith("metalearners"):
            var_list = estimand.treatment_variable + self._effect_modifier_names
            expr += "+".join(var_list)
        else:
            var_list = estimand.treatment_variable + self._observed_common_causes_names
            expr += "+".join(var_list)
            expr += " | " + ",".join(self._effect_modifier_names)
        return expr

    def shap_values(self, df: pd.DataFrame, *args, **kwargs):
        return self.estimator.shap_values(df[self._effect_modifier_names].values, *args, **kwargs)

    def apply_multitreatment(self, df: pd.DataFrame, fun: Callable, *args, **kwargs):
        ests = []
        assert not isinstance(self._treatment_value, str)
        assert is_sequence(self._treatment_value)

        if df is None:
            filtered_df = None
        else:
            filtered_df = df[self._effect_modifier_names].values

        for tv in self._treatment_value:
            ests.append(
                fun(
                    filtered_df,
                    T0=self._control_value,
                    T1=tv,
                    *args,
                    **kwargs,
                )
            )
        est = np.stack(ests, axis=1)
        return est

    def effect(self, df: pd.DataFrame, *args, **kwargs) -> np.ndarray:
        """
        Pointwise estimated treatment effect,
        output shape n_units x n_treatment_values (not counting control)
        :param df: Features of the units to evaluate
        :param args: passed through to the underlying estimator
        :param kwargs: passed through to the underlying estimator
        """

        def effect_fun(filtered_df, T0, T1, *args, **kwargs):
            return self.estimator.effect(filtered_df, T0=T0, T1=T1, *args, **kwargs)

        return self.apply_multitreatment(df, effect_fun, *args, **kwargs)

    def effect_interval(self, df: pd.DataFrame, *args, **kwargs) -> np.ndarray:
        """
        Pointwise confidence intervals for the estimated treatment effect
        :param df: Features of the units to evaluate
        :param args: passed through to the underlying estimator
        :param kwargs: passed through to the underlying estimator
        """

        def effect_interval_fun(filtered_df, T0, T1, *args, **kwargs):
            return self.estimator.effect_interval(
                filtered_df, T0=T0, T1=T1, alpha=1 - self.confidence_level, *args, **kwargs
            )

        return self.apply_multitreatment(df, effect_interval_fun, *args, **kwargs)

    def effect_inference(self, df: pd.DataFrame, *args, **kwargs):
        """
        Inference (uncertainty) results produced by the underlying EconML estimator
        :param df: Features of the units to evaluate
        :param args: passed through to the underlying estimator
        :param kwargs: passed through to the underlying estimator
        """

        def effect_inference_fun(filtered_df, T0, T1, *args, **kwargs):
            return self.estimator.effect_inference(filtered_df, T0=T0, T1=T1, *args, **kwargs)

        return self.apply_multitreatment(df, effect_inference_fun, *args, **kwargs)

    def effect_tt(self, df: pd.DataFrame, *args, **kwargs):
        """
        Effect of the actual treatment that was applied to each unit
        ("effect of Treatment on the Treated")
        :param df: Features of the units to evaluate
        :param args: passed through to estimator.effect()
        :param kwargs: passed through to estimator.effect()
        """

        eff = self.effect(df, *args, **kwargs).reshape((len(df), len(self._treatment_value)))

        out = np.zeros(len(df))
        treatment_value = parse_state(self._treatment_value)
        treatment_name = parse_state(self._treatment_name)[0]

        eff = np.reshape(eff, (len(df), len(treatment_value)))

        # For each unit, return the estimated effect of the treatment value
        # that was actually applied to the unit
        for c, col in enumerate(treatment_value):
            out[df[treatment_name] == col] = eff[df[treatment_name] == col, c]
        return pd.Series(data=out, index=df.index)



import itertools
from typing import List, Optional, Union

import pandas as pd
import statsmodels.api as sm

from dowhy.causal_estimator import CausalEstimator
from dowhy.causal_estimators.regression_estimator import RegressionEstimator
from dowhy.causal_identifier import IdentifiedEstimand


class LinearRegressionEstimator(RegressionEstimator):
    """Compute effect of treatment using linear regression.

    Fits a regression model for estimating the outcome using treatment(s) and confounders. For a univariate treatment, the treatment effect is equivalent to the coefficient of the treatment variable.

    Simple method to show the implementation of a causal inference method that can handle multiple treatments and heterogeneity in treatment. Requires a strong assumption that all relationships from (T, W) to Y are linear.

    """

    def __init__(
        self,
        identified_estimand: IdentifiedEstimand,
        test_significance: bool = False,
        evaluate_effect_strength: bool = False,
        confidence_intervals: bool = False,
        num_null_simulations: int = CausalEstimator.DEFAULT_NUMBER_OF_SIMULATIONS_STAT_TEST,
        num_simulations: int = CausalEstimator.DEFAULT_NUMBER_OF_SIMULATIONS_CI,
        sample_size_fraction: int = CausalEstimator.DEFAULT_SAMPLE_SIZE_FRACTION,
        confidence_level: float = CausalEstimator.DEFAULT_CONFIDENCE_LEVEL,
        need_conditional_estimates: Union[bool, str] = "auto",
        num_quantiles_to_discretize_cont_cols: int = CausalEstimator.NUM_QUANTILES_TO_DISCRETIZE_CONT_COLS,
        **kwargs,
    ):
        """
        :param identified_estimand: probability expression
            representing the target identified estimand to estimate.
        :param test_significance: Binary flag or a string indicating whether to test significance and by which method. All estimators support test_significance="bootstrap" that estimates a p-value for the obtained estimate using the bootstrap method. Individual estimators can override this to support custom testing methods. The bootstrap method supports an optional parameter, num_null_simulations. If False, no testing is done. If True, significance of the estimate is tested using the custom method if available, otherwise by bootstrap.
        :param evaluate_effect_strength: (Experimental) whether to evaluate the strength of effect
        :param confidence_intervals: Binary flag or a string indicating whether the confidence intervals should be computed and which method should be used. All methods support estimation of confidence intervals using the bootstrap method by using the parameter confidence_intervals="bootstrap". The bootstrap method takes in two arguments (num_simulations and sample_size_fraction) that can be optionally specified in the params dictionary. Estimators may also override this to implement their own confidence interval method. If this parameter is False, no confidence intervals are computed. If True, confidence intervals are computed by the estimator's specific method if available, otherwise through bootstrap
        :param num_null_simulations: The number of simulations for testing the
            statistical significance of the estimator
        :param num_simulations: The number of simulations for finding the
            confidence interval (and/or standard error) for a estimate
        :param sample_size_fraction: The size of the sample for the bootstrap
            estimator
        :param confidence_level: The confidence level of the confidence
            interval estimate
        :param need_conditional_estimates: Boolean flag indicating whether
            conditional estimates should be computed. Defaults to True if
            there are effect modifiers in the graph
        :param num_quantiles_to_discretize_cont_cols: The number of quantiles
            into which a numeric effect modifier is split, to enable
            estimation of conditional treatment effect over it.
        :param kwargs: (optional) Additional estimator-specific parameters
        """
        super().__init__(
            identified_estimand=identified_estimand,
            test_significance=test_significance,
            evaluate_effect_strength=evaluate_effect_strength,
            confidence_intervals=confidence_intervals,
            num_null_simulations=num_null_simulations,
            num_simulations=num_simulations,
            sample_size_fraction=sample_size_fraction,
            confidence_level=confidence_level,
            need_conditional_estimates=need_conditional_estimates,
            num_quantiles_to_discretize_cont_cols=num_quantiles_to_discretize_cont_cols,
            **kwargs,
        )
        self.logger.info("INFO: Using Linear Regression Estimator")
        self._linear_model = self.model

    def fit(
        self,
        data: pd.DataFrame,
        treatment_name: str,
        outcome_name: str,
        effect_modifier_names: Optional[List[str]] = None,
    ):
        """
        Fits the estimator with data for effect estimation
        :param data: data frame containing the data
        :param treatment: name of the treatment variable
        :param outcome: name of the outcome variable
        :param effect_modifiers: Variables on which to compute separate
                    effects, or return a heterogeneous effect function. Not all
                    methods support this currently.
        """
        return super().fit(data, treatment_name, outcome_name, effect_modifier_names=effect_modifier_names)

    def construct_symbolic_estimator(self, estimand):
        expr = "b: " + ",".join(estimand.outcome_variable) + "~"
        var_list = estimand.treatment_variable + estimand.get_backdoor_variables()
        expr += "+".join(var_list)
        if self._effect_modifier_names:
            interaction_terms = [
                "{0}*{1}".format(x[0], x[1])
                for x in itertools.product(estimand.treatment_variable, self._effect_modifier_names)
            ]
            expr += "+" + "+".join(interaction_terms)
        return expr

    def predict_fn(self, model, features):
        return model.predict(features)

    def _build_model(self):
        features = self._build_features()
        model = sm.OLS(self._outcome, features).fit()
        return (features, model)

    def _estimate_confidence_intervals(self, confidence_level, method=None):
        if self._effect_modifier_names:
            # The average treatment effect is a combination of different
            # regression coefficients. Complicated to compute the confidence
            # interval analytically. For example, if y=a + b1.t + b2.tx, then
            # the average treatment effect is b1+b2.mean(x).
            # Refer Gelman, Hill. ARM Book. Chapter 9
            # http://www.stat.columbia.edu/~gelman/arm/chap9.pdf
            # TODO: Looking for contributions
            raise NotImplementedError
        else:
            conf_ints = self.model.conf_int(alpha=1 - confidence_level)
            # For a linear regression model, the causal effect of a variable is equal to the coefficient corresponding to the
            # variable. Hence, the model by default outputs the confidence interval corresponding to treatment=1 and control=0.
            # So for custom treatment and control values, we must multiply the confidence interval by the difference of the two.
            return (self._treatment_value - self._control_value) * conf_ints.to_numpy()[
                1 : (len(self._treatment_name) + 1), :
            ]

    def _estimate_std_error(self, method=None):
        if self._effect_modifier_names:
            raise NotImplementedError
        else:
            std_error = self.model.bse[1 : (len(self._treatment_name) + 1)]

            # For a linear regression model, the causal effect of a variable is equal to the coefficient corresponding to the
            # variable. Hence, the model by default outputs the standard error corresponding to treatment=1 and control=0.
            # So for custom treatment and control values, we must multiply the standard error by the difference of the two.
            return (self._treatment_value - self._control_value) * std_error.to_numpy()

    def _test_significance(self, estimate_value, method=None):
        pvalue = self.model.pvalues[1 : (len(self._treatment_name) + 1)]
        return {"p_value": pvalue.to_numpy()}



import numpy as np
import pytest
from flaky import flaky

from dowhy.causal_estimator import CausalEstimate
from dowhy.causal_identifier.identified_estimand import IdentifiedEstimand
from dowhy.causal_refuter import CausalRefuter


class MockRefuter(CausalRefuter):
    pass


def test_causal_refuter_placeholder_method():
    refuter = MockRefuter(None, IdentifiedEstimand(None, None, None), None)
    with pytest.raises(NotImplementedError):
        refuter.refute_estimate()


@flaky(max_runs=3)
def test_causal_refuter_bootstrap_test():
    estimator = CausalEstimate(0, None, None, None, None)
    refuter = MockRefuter(None, IdentifiedEstimand(None, None, None), None)
    simulations = np.random.normal(0, 1, 5000)
    pvalue = refuter.perform_bootstrap_test(estimator, simulations)
    assert pvalue > 0.95



import inspect
from importlib import import_module
from typing import Any, List, Optional, Protocol, Union
from warnings import warn

import pandas as pd

from dowhy.causal_estimator import CausalEstimate, CausalEstimator
from dowhy.causal_identifier import IdentifiedEstimand


class _CausalmlEstimator(Protocol):
    def estimate_ate(self, *args, **kwargs):
        ...

    def fit_predict(self, *args, **kwargs):
        ...


class Causalml(CausalEstimator):
    """Wrapper class for estimators from the causalml library.

    Supports additional parameters as listed below. For specific
    parameters of each estimator, refer to the CausalML docs.

    """

    def __init__(
        self,
        identified_estimand: IdentifiedEstimand,
        causalml_estimator: Union[_CausalmlEstimator, str],
        test_significance: bool = False,
        evaluate_effect_strength: bool = False,
        confidence_intervals: bool = False,
        num_null_simulations: int = CausalEstimator.DEFAULT_NUMBER_OF_SIMULATIONS_STAT_TEST,
        num_simulations: int = CausalEstimator.DEFAULT_NUMBER_OF_SIMULATIONS_CI,
        sample_size_fraction: int = CausalEstimator.DEFAULT_SAMPLE_SIZE_FRACTION,
        confidence_level: float = CausalEstimator.DEFAULT_CONFIDENCE_LEVEL,
        need_conditional_estimates: Union[bool, str] = "auto",
        num_quantiles_to_discretize_cont_cols: int = CausalEstimator.NUM_QUANTILES_TO_DISCRETIZE_CONT_COLS,
        **kwargs,
    ):
        """
        :param identified_estimand: probability expression
            representing the target identified estimand to estimate.
        :param causalml_methodname: Fully qualified name of causalml estimator
            class.
        :param test_significance: Binary flag or a string indicating whether to test significance and by which method. All estimators support test_significance="bootstrap" that estimates a p-value for the obtained estimate using the bootstrap method. Individual estimators can override this to support custom testing methods. The bootstrap method supports an optional parameter, num_null_simulations. If False, no testing is done. If True, significance of the estimate is tested using the custom method if available, otherwise by bootstrap.
        :param evaluate_effect_strength: (Experimental) whether to evaluate the strength of effect
        :param confidence_intervals: Binary flag or a string indicating whether the confidence intervals should be computed and which method should be used. All methods support estimation of confidence intervals using the bootstrap method by using the parameter confidence_intervals="bootstrap". The bootstrap method takes in two arguments (num_simulations and sample_size_fraction) that can be optionally specified in the params dictionary. Estimators may also override this to implement their own confidence interval method. If this parameter is False, no confidence intervals are computed. If True, confidence intervals are computed by the estimator's specific method if available, otherwise through bootstrap
        :param num_null_simulations: The number of simulations for testing the
            statistical significance of the estimator
        :param num_simulations: The number of simulations for finding the
            confidence interval (and/or standard error) for a estimate
        :param sample_size_fraction: The size of the sample for the bootstrap
            estimator
        :param confidence_level: The confidence level of the confidence
            interval estimate
        :param need_conditional_estimates: Boolean flag indicating whether
            conditional estimates should be computed. Defaults to True if
            there are effect modifiers in the graph
        :param num_quantiles_to_discretize_cont_cols: The number of quantiles
            into which a numeric effect modifier is split, to enable
            estimation of conditional treatment effect over it.
        :param kwargs: (optional) Additional estimator-specific parameters
        """
        super().__init__(
            identified_estimand=identified_estimand,
            test_significance=test_significance,
            evaluate_effect_strength=evaluate_effect_strength,
            confidence_intervals=confidence_intervals,
            num_null_simulations=num_null_simulations,
            num_simulations=num_simulations,
            sample_size_fraction=sample_size_fraction,
            confidence_level=confidence_level,
            need_conditional_estimates=need_conditional_estimates,
            num_quantiles_to_discretize_cont_cols=num_quantiles_to_discretize_cont_cols,
            causalml_estimator=causalml_estimator,
            **kwargs,
        )
        # Add the identification method used in the estimator
        self.identifier_method = self._target_estimand.identifier_method
        self.logger.debug("The identifier method used {}".format(self.identifier_method))

        if isinstance(causalml_estimator, str):
            warn(
                "Using a string to specify the value for causalml_estimator is now deprecated, please provide an instance of a causalml estimator object",
                DeprecationWarning,
                stacklevel=2,
            )
            estimator_class = self._get_causalml_class_object(causalml_estimator)
            self.estimator = estimator_class(**kwargs["init_params"])
        else:
            self.estimator = causalml_estimator
        self.logger.info("INFO: Using CausalML Estimator")

    def fit(
        self,
        data: pd.DataFrame,
        treatment_name: str,
        outcome_name: str,
        effect_modifier_names: Optional[List[str]] = None,
    ):
        """
        Fits the estimator with data for effect estimation
        :param data: data frame containing the data
        :param treatment: name of the treatment variable
        :param outcome: name of the outcome variable
        :param effect_modifiers: Variables on which to compute separate
                    effects, or return a heterogeneous effect function. Not all
                    methods support this currently.
        """
        self._set_data(data, treatment_name, outcome_name)
        self._set_effect_modifiers(effect_modifier_names)

        # Check the backdoor variables being used
        self.logger.debug("Back-door variables used:" + ",".join(self._target_estimand.get_backdoor_variables()))

        # Add the observed confounders and one hot encode the categorical variables
        self._observed_common_causes_names = self._target_estimand.get_backdoor_variables()
        if self._observed_common_causes_names:
            # Get the data of the unobserved confounders
            self._observed_common_causes = self._data[self._observed_common_causes_names]
            # One hot encode the data if they are categorical
            self._observed_common_causes = pd.get_dummies(self._observed_common_causes, drop_first=True)
        else:
            self._observed_common_causes = []

        # Check the instrumental variables involved
        self.logger.debug("Instrumental variables used:" + ",".join(self._target_estimand.instrumental_variables))

        # Perform the same actions as the above
        self._instrumental_variable_names = self._target_estimand.instrumental_variables
        if self._instrumental_variable_names:
            self._instrumental_variables = self._data[self._instrumental_variable_names]
            self._instrumental_variables = pd.get_dummies(self._instrumental_variables, drop_first=True)
        else:
            self._instrumental_variables = []

        self.symbolic_estimator = self.construct_symbolic_estimator(self._target_estimand)
        self.logger.info(self.symbolic_estimator)

        return self

    def _get_causalml_class_object(self, module_method_name, *args, **kwargs):

        try:
            (module_name, _, class_name) = module_method_name.rpartition(".")
            estimator_module = import_module(module_name)
            estimator_class = getattr(estimator_module, class_name)

        except (AttributeError, AssertionError, ImportError):
            raise ImportError(
                "Error loading {}.{}. Double-check the method name and ensure that all causalml dependencies are installed.".format(
                    module_name, class_name
                )
            )
        return estimator_class

    def estimate_effect(
        self, data: pd.DataFrame = None, treatment_value: Any = 1, control_value: Any = 0, target_units=None, **_
    ):
        """
        data: dataframe containing the data on which treatment effect is to be estimated.
        treatment_value: value of the treatment variable for which the effect is to be estimated.
        control_value: value of the treatment variable that denotes its absence (usually 0)
        target_units: The units for which the treatment effect should be estimated.
                     It can be a DataFrame that contains values of the effect_modifiers and effect will be estimated only for this new data.
                     It can also be a lambda function that can be used as an index for the data (pandas DataFrame) to select the required rows.
        """
        if data is None:
            data = self._data

        self._target_units = target_units
        self._treatment_value = treatment_value
        self._control_value = control_value
        X_names = self._observed_common_causes_names + self._effect_modifier_names

        # Both the outcome and the treatment have to be 1D arrays according to the CausalML API
        y_name = self._outcome_name
        treatment_name = self._treatment_name[0]  # As we have only one treatment variable
        # We want to pass 'v0' rather than ['v0'] to prevent a shape mismatch

        func_args = {"X": data[X_names], "y": data[y_name], "treatment": data[treatment_name]}

        arg_names = inspect.getfullargspec(self.estimator.estimate_ate)[0]
        matched_args = {arg: func_args[arg] for arg in func_args.keys() if arg in arg_names}
        self.logger.debug(matched_args)
        value_tuple = self.estimator.estimate_ate(**matched_args)

        # For CATEs
        arg_names = inspect.getfullargspec(self.estimator.fit_predict)[0]
        matched_args = {arg: func_args[arg] for arg in func_args.keys() if arg in arg_names}
        cate_estimates = self.estimator.fit_predict(**matched_args)

        estimate = CausalEstimate(
            estimate=value_tuple[0],
            control_value=control_value,
            treatment_value=treatment_value,
            target_estimand=self._target_estimand,
            realized_estimand_expr=self.symbolic_estimator,
            cate_estimates=cate_estimates,
            effect_intervals=(value_tuple[1], value_tuple[2]),
            _estimator_object=self.estimator,
        )

        estimate.add_estimator(self)
        return estimate

    def construct_symbolic_estimator(self, estimand):
        expr = "b: " + ",".join(estimand.outcome_variable) + "~"
        # TODO we are conditioning on a postive treatment
        # TODO create an expression corresponding to each estimator used
        var_list = estimand.treatment_variable + estimand.get_backdoor_variables()
        expr += "+".join(var_list)
        return expr



import pdb

import numpy as np
import pytest
from pytest import mark

from .base import TestRefuter


def simple_linear_outcome_model(X_train, output_train):
    # The outcome is a linear function of the confounder
    # The slope is 1,2 and the intercept is 3
    return lambda X_train: X_train[:, 0] + 2 * X_train[:, 1] + 3


@mark.usefixtures("fixed_seed")
class TestDummyOutcomeRefuter(object):
    @mark.parametrize(["error_tolerence", "estimator_method"], [(0.03, "iv.instrumental_variable")])
    def test_refutation_dummy_outcome_refuter_default_continuous_treatment(self, error_tolerence, estimator_method):
        refuter_tester = TestRefuter(error_tolerence, estimator_method, "dummy_outcome_refuter")
        refuter_tester.continuous_treatment_testsuite(num_dummyoutcome_simulations=100)

    @mark.parametrize(
        ["error_tolerence", "estimator_method", "num_samples"], [(0.1, "backdoor.propensity_score_matching", 1000)]
    )
    def test_refutation_dummy_outcome_refuter_default_binary_treatment(
        self, error_tolerence, estimator_method, num_samples
    ):
        refuter_tester = TestRefuter(error_tolerence, estimator_method, "dummy_outcome_refuter")
        refuter_tester.binary_treatment_testsuite(tests_to_run="atleast-one-common-cause", num_samples=num_samples)

    @mark.parametrize(
        ["error_tolerence", "estimator_method", "transformations"],
        [(0.05, "iv.instrumental_variable", [("zero", ""), ("noise", {"std_dev": 1})])],
    )
    def test_refutation_dummy_outcome_refuter_randomly_generated_continuous_treatment(
        self, error_tolerence, estimator_method, transformations
    ):
        refuter_tester = TestRefuter(
            error_tolerence, estimator_method, "dummy_outcome_refuter", transformations=transformations
        )

        refuter_tester.continuous_treatment_testsuite()

    @mark.parametrize(
        ["error_tolerence", "estimator_method", "transformations", "num_samples"],
        [(0.05, "backdoor.propensity_score_matching", [("zero", ""), ("noise", {"std_dev": 1})], 1000)],
    )
    def test_refutation_dummy_outcome_refuter_randomly_generated_binary_treatment(
        self, error_tolerence, estimator_method, transformations, num_samples
    ):
        refuter_tester = TestRefuter(
            error_tolerence, estimator_method, "dummy_outcome_refuter", transformations=transformations
        )

        refuter_tester.binary_treatment_testsuite(tests_to_run="atleast-one-common-cause", num_samples=num_samples)

    @mark.parametrize(
        ["error_tolerence", "estimator_method", "transformations"],
        [(0.03, "iv.instrumental_variable", [("permute", {"permute_fraction": 1})])],
    )
    def test_refutation_dummy_outcome_refuter_permute_data_continuous_treatment(
        self, error_tolerence, estimator_method, transformations
    ):
        refuter_tester = TestRefuter(
            error_tolerence, estimator_method, "dummy_outcome_refuter", transformations=transformations
        )

        refuter_tester.continuous_treatment_testsuite()

    @mark.parametrize(
        ["error_tolerence", "estimator_method", "transformations", "num_samples"],
        [(0.1, "backdoor.linear_regression", [("permute", {"permute_fraction": 1})], 1000)],
    )
    def test_refutation_dummy_outcome_refuter_permute_data_binary_treatment(
        self, error_tolerence, estimator_method, transformations, num_samples
    ):
        refuter_tester = TestRefuter(
            error_tolerence, estimator_method, "dummy_outcome_refuter", transformations=transformations
        )

        refuter_tester.binary_treatment_testsuite(tests_to_run="atleast-one-common-cause", num_samples=num_samples)

    @mark.parametrize(
        ["error_tolerence", "estimator_method", "transformations"],
        [(0.2, "iv.instrumental_variable", [(simple_linear_outcome_model, {}), ("noise", {"std_dev": 1})])],
    )
    def test_refutation_dummy_outcome_refuter_custom_function_linear_regression_with_noise_continuous_treatment(
        self, error_tolerence, estimator_method, transformations
    ):
        refuter_tester = TestRefuter(
            error_tolerence, estimator_method, "dummy_outcome_refuter", transformations=transformations
        )
        refuter_tester.continuous_treatment_testsuite(tests_to_run="atleast-one-common-cause")

    @mark.xfail
    @mark.parametrize(
        ["error_tolerence", "estimator_method", "transformations", "num_samples"],
        [(0.2, "backdoor.linear_regression", [(simple_linear_outcome_model, {}), ("noise", {"std_dev": 1})], 1000)],
    )
    def test_refutation_dummy_outcome_refuter_custom_function_linear_regression_with_noise_binary_treatment(
        self, error_tolerence, estimator_method, transformations, num_samples
    ):
        refuter_tester = TestRefuter(
            error_tolerence, estimator_method, "dummy_outcome_refuter", transformations=transformations
        )
        refuter_tester.binary_treatment_testsuite(tests_to_run="atleast-one-common-cause", num_samples=num_samples)

    @mark.parametrize(
        ["error_tolerence", "estimator_method", "transformations"],
        [
            (
                0.2,
                "backdoor.linear_regression",
                [("permute", {"permute_fraction": 0.5}), (simple_linear_outcome_model, {}), ("noise", {"std_dev": 1})],
            )
        ],
    )
    def test_refutation_dummy_outcome_refuter_custom_function_linear_regression_with_permute_continuous_treatment(
        self, error_tolerence, estimator_method, transformations
    ):
        refuter_tester = TestRefuter(
            error_tolerence, estimator_method, "dummy_outcome_refuter", transformations=transformations
        )
        refuter_tester.continuous_treatment_testsuite(tests_to_run="atleast-one-common-cause")

    @mark.xfail
    @mark.parametrize(
        ["error_tolerence", "estimator_method", "transformations", "num_samples"],
        [
            (
                0.2,
                "backdoor.propensity_score_matching",
                [("permute", {"permute_fraction": 0.5}), (simple_linear_outcome_model, {}), ("noise", {"std_dev": 1})],
                1000,
            )
        ],
    )
    def test_refutation_dummy_outcome_refuter_custom_function_linear_regression_with_permute_binary_treatment(
        self, error_tolerence, estimator_method, transformations, num_samples
    ):
        refuter_tester = TestRefuter(
            error_tolerence, estimator_method, "dummy_outcome_refuter", transformations=transformations
        )
        refuter_tester.binary_treatment_testsuite(tests_to_run="atleast-one-common-cause", num_samples=num_samples)

    @mark.parametrize(
        ["error_tolerence", "estimator_method", "transformations"],
        [(0.01, "iv.instrumental_variable", [("linear_regression", {}), ("zero", ""), ("noise", {"std_dev": 1})])],
    )
    def test_refutation_dummy_outcome_refuter_internal_linear_regression_continuous_treatment(
        self, error_tolerence, estimator_method, transformations
    ):
        refuter_tester = TestRefuter(
            error_tolerence, estimator_method, "dummy_outcome_refuter", transformations=transformations
        )
        refuter_tester.continuous_treatment_testsuite(tests_to_run="atleast-one-common-cause")

    @mark.parametrize(
        ["error_tolerence", "estimator_method", "transformations", "num_samples"],
        [
            (
                0.2,
                "backdoor.propensity_score_matching",
                [("linear_regression", {}), ("zero", ""), ("noise", {"std_dev": 1})],
                1000,
            )
        ],
    )
    def test_refutation_dummy_outcome_refuter_internal_linear_regression_binary_treatment(
        self, error_tolerence, estimator_method, transformations, num_samples
    ):
        refuter_tester = TestRefuter(
            error_tolerence, estimator_method, "dummy_outcome_refuter", transformations=transformations
        )
        refuter_tester.binary_treatment_testsuite(tests_to_run="atleast-one-common-cause", num_samples=num_samples)

    @mark.parametrize(
        ["error_tolerence", "estimator_method", "transformations"],
        [(0.2, "iv.instrumental_variable", [("knn", {"n_neighbors": 5}), ("zero", ""), ("noise", {"std_dev": 1})])],
    )
    def test_refutation_dummy_outcome_refuter_internal_knn_continuous_treatment(
        self, error_tolerence, estimator_method, transformations
    ):
        refuter_tester = TestRefuter(
            error_tolerence, estimator_method, "dummy_outcome_refuter", transformations=transformations
        )
        refuter_tester.continuous_treatment_testsuite(tests_to_run="atleast-one-common-cause")

    @mark.parametrize(
        ["error_tolerence", "estimator_method", "transformations", "num_samples"],
        [
            (
                0.2,
                "backdoor.propensity_score_matching",
                [("knn", {"n_neighbors": 5}), ("zero", ""), ("noise", {"std_dev": 1})],
                1000,
            )
        ],
    )
    def test_refutation_dummy_outcome_refuter_internal_knn_binary_treatment(
        self, error_tolerence, estimator_method, transformations, num_samples
    ):
        refuter_tester = TestRefuter(
            error_tolerence, estimator_method, "dummy_outcome_refuter", transformations=transformations
        )
        refuter_tester.binary_treatment_testsuite(tests_to_run="atleast-one-common-cause", num_samples=num_samples)

    @mark.parametrize(
        ["error_tolerence", "estimator_method", "transformations", "num_samples"],
        [
            (
                0.01,
                "iv.instrumental_variable",
                [("svm", {"C": 1, "epsilon": 0.2}), ("zero", ""), ("noise", {"std_dev": 1})],
                10000,
            )
        ],
    )
    def test_refutation_dummy_outcome_refuter_internal_svm_continuous_treatment(
        self, error_tolerence, estimator_method, transformations, num_samples
    ):
        refuter_tester = TestRefuter(
            error_tolerence, estimator_method, "dummy_outcome_refuter", transformations=transformations
        )
        refuter_tester.continuous_treatment_testsuite(num_samples=num_samples, tests_to_run="atleast-one-common-cause")

    @mark.parametrize(
        ["error_tolerence", "estimator_method", "transformations", "num_samples"],
        [
            (
                0.1,
                "backdoor.propensity_score_matching",
                [("svm", {"C": 1, "epsilon": 0.2}), ("zero", ""), ("noise", {"std_dev": 1})],
                1000,
            )
        ],
    )
    def test_refutation_dummy_outcome_refuter_internal_svm_binary_treatment(
        self, error_tolerence, estimator_method, transformations, num_samples
    ):
        refuter_tester = TestRefuter(
            error_tolerence, estimator_method, "dummy_outcome_refuter", transformations=transformations
        )
        refuter_tester.binary_treatment_testsuite(num_samples=num_samples, tests_to_run="atleast-one-common-cause")

    @mark.parametrize(
        ["error_tolerence", "estimator_method", "transformations", "num_samples"],
        [
            (
                0.01,
                "iv.instrumental_variable",
                [("random_forest", {"max_depth": 20}), ("zero", ""), ("noise", {"std_dev": 1})],
                10000,
            )
        ],
    )
    def test_refutation_dummy_outcome_refuter_internal_random_forest_continuous_treatment(
        self, error_tolerence, estimator_method, transformations, num_samples
    ):
        refuter_tester = TestRefuter(
            error_tolerence, estimator_method, "dummy_outcome_refuter", transformations=transformations
        )
        refuter_tester.continuous_treatment_testsuite(num_samples, tests_to_run="atleast-one-common-cause")

    @mark.parametrize(
        ["error_tolerence", "estimator_method", "transformations", "num_samples"],
        [
            (
                0.1,
                "backdoor.propensity_score_matching",
                [("random_forest", {"max_depth": 20}), ("zero", ""), ("noise", {"std_dev": 1})],
                1000,
            )
        ],
    )
    def test_refutation_dummy_outcome_refuter_internal_random_forest_binary_treatment(
        self, error_tolerence, estimator_method, transformations, num_samples
    ):
        refuter_tester = TestRefuter(
            error_tolerence, estimator_method, "dummy_outcome_refuter", transformations=transformations
        )
        refuter_tester.binary_treatment_testsuite(num_samples, tests_to_run="atleast-one-common-cause")

    # As we run with only one common cause and one instrument variable we run with (?, 2)
    @mark.parametrize(
        ["error_tolerence", "estimator_method", "transformations"],
        [
            (
                0.01,
                "iv.instrumental_variable",
                [
                    ("neural_network", {"solver": "lbfgs", "alpha": 1e-5, "hidden_layer_sizes": (5, 2)}),
                    ("zero", ""),
                    ("noise", {"std_dev": 1}),
                ],
            )
        ],
    )
    def test_refutation_dummy_outcome_refuter_internal_neural_network_continuous_treatment(
        self, error_tolerence, estimator_method, transformations
    ):
        refuter_tester = TestRefuter(
            error_tolerence, estimator_method, "dummy_outcome_refuter", transformations=transformations
        )
        refuter_tester.continuous_treatment_testsuite(tests_to_run="atleast-one-common-cause")

    @mark.parametrize(
        ["error_tolerence", "estimator_method", "transformations", "num_samples"],
        [
            (
                0.1,
                "backdoor.propensity_score_matching",
                [
                    ("neural_network", {"solver": "lbfgs", "alpha": 1e-5, "hidden_layer_sizes": (5, 2)}),
                    ("zero", ""),
                    ("noise", {"std_dev": 1}),
                ],
                1000,
            )
        ],
    )
    def test_refutation_dummy_outcome_refuter_internal_neural_network_binary_treatment(
        self, error_tolerence, estimator_method, transformations, num_samples
    ):
        refuter_tester = TestRefuter(
            error_tolerence, estimator_method, "dummy_outcome_refuter", transformations=transformations
        )
        refuter_tester.binary_treatment_testsuite(num_samples=num_samples, tests_to_run="atleast-one-common-cause")



import logging

from dowhy.causal_identifier import EstimandType, identify_effect, identify_effect_auto, identify_effect_id
from dowhy.causal_model import CausalModel

logging.getLogger(__name__).addHandler(logging.NullHandler())

#
# 0.0.0 is standard placeholder for poetry-dynamic-versioning
# any changes to this should not be checked in
#
__version__ = "0.9"


__all__ = ["EstimandType", "identify_effect_auto", "identify_effect_id", "identify_effect", "CausalModel"]



"""Functions in this module should be considered experimental, meaning there might be breaking API changes in the
future.
"""

from typing import Callable, List, Optional, Union

import numpy as np
import scipy
from causallearn.utils.KCI.KCI import KCI_CInd, KCI_UInd
from joblib import Parallel, delayed
from sklearn.preprocessing import scale

import dowhy.gcm.config as config
from dowhy.gcm.independence_test.kernel_operation import approximate_rbf_kernel_features
from dowhy.gcm.stats import quantile_based_fwer
from dowhy.gcm.util.general import apply_one_hot_encoding, fit_one_hot_encoders, set_random_seed, shape_into_2d


def kernel_based(
    X: np.ndarray,
    Y: np.ndarray,
    Z: Optional[np.ndarray] = None,
    use_bootstrap: bool = True,
    bootstrap_num_runs: int = 10,
    bootstrap_num_samples_per_run: int = 2000,
    bootstrap_n_jobs: Optional[int] = None,
    p_value_adjust_func: Callable[[Union[np.ndarray, List[float]]], float] = quantile_based_fwer,
    **kwargs,
) -> float:
    """Prepares the data and uses kernel (conditional) independence test. The independence test estimates a p-value
    for the null hypothesis that X and Y are independent (given Z). Depending whether Z is given, a conditional or
    pairwise independence test is performed.

    Here, we utilize the implementations of the https://github.com/cmu-phil/causal-learn package.

    If Z is given: Using KCI as conditional independence test, i.e. we use https://github.com/cmu-phil/causal-learn/blob/main/causallearn/utils/KCI/KCI.py#L238.
    If Z is not given: Using KCI as pairwise independence test, i.e. we use https://github.com/cmu-phil/causal-learn/blob/main/causallearn/utils/KCI/KCI.py#L17.

    Note:
    - The data can be multivariate, i.e. the given input matrices can have multiple columns.
    - Categorical data need to be represented as strings.

    Based on the work:
    - K. Zhang, J. Peters, D. Janzing, B. Schölkopf. *Kernel-based Conditional Independence Test and Application in Causal Discovery*. UAI'11, Pages 804–813, 2011.
    - A. Gretton, K. Fukumizu, C.-H. Teo, L. Song, B. Schölkopf, A. Smola. *A Kernel Statistical Test of Independence*. NIPS 21, 2007.

    For more information about configuring the kernel independence test, see:
    - https://github.com/cmu-phil/causal-learn/blob/main/causallearn/utils/KCI/KCI.py#L17 (if Z is not given)
    - https://github.com/cmu-phil/causal-learn/blob/main/causallearn/utils/KCI/KCI.py#L238 (if Z is given)

    :param X: Data matrix for observations from X.
    :param Y: Data matrix for observations from Y.
    :param Z: Optional data matrix for observations from Z. This is the conditional variable.
    :param use_bootstrap: If True, the independence tests are performed on multiple subsets of the data and the final
                          p-value is constructed based on the provided p_value_adjust_func function.
    :param bootstrap_num_runs: Number of bootstrap runs (only relevant if use_bootstrap is True).
    :param bootstrap_num_samples_per_run: Number of samples used in a bootstrap run (only relevant if use_bootstrap is
                                          True).
    :param bootstrap_n_jobs: Number of parallel jobs for the bootstrap runs.
    :param p_value_adjust_func: A callable that expects a numpy array of multiple p-values and returns one p-value. This
                                is typically used a family wise error rate control method.
    :return: The p-value for the null hypothesis that X and Y are independent (given Z).
    """
    bootstrap_n_jobs = config.default_n_jobs if bootstrap_n_jobs is None else bootstrap_n_jobs

    X = _remove_constant_columns(X)
    Y = _remove_constant_columns(Y)

    if X.shape[1] == 0 or Y.shape[1] == 0:
        # Either X and/or Y is constant.
        return 1.0

    if Z is not None:
        Z = _remove_constant_columns(Z)
        if Z.shape[1] == 0:
            # If Z is empty, we are in the pairwise setting.
            Z = None

    if "est_width" not in kwargs:
        kwargs["est_width"] = "median"

    def evaluate_kernel_test_on_samples(
        X: np.ndarray, Y: np.ndarray, Z: np.ndarray, parallel_random_seed: int
    ) -> float:
        set_random_seed(parallel_random_seed)

        if Z is None:
            X, Y = _convert_to_numeric(*shape_into_2d(X, Y))
            return KCI_UInd(**kwargs).compute_pvalue(X, Y)[0]
        else:
            X, Y, Z = _convert_to_numeric(*shape_into_2d(X, Y, Z))
            return KCI_CInd(**kwargs).compute_pvalue(X, Y, Z)[0]

    if use_bootstrap and X.shape[0] > bootstrap_num_samples_per_run:
        random_indices = [
            np.random.choice(X.shape[0], min(X.shape[0], bootstrap_num_samples_per_run), replace=False)
            for run in range(bootstrap_num_runs)
        ]

        random_seeds = np.random.randint(np.iinfo(np.int32).max, size=len(random_indices))
        p_values = Parallel(n_jobs=bootstrap_n_jobs)(
            delayed(evaluate_kernel_test_on_samples)(
                X[indices], Y[indices], Z[indices] if Z is not None else None, random_seed
            )
            for indices, random_seed in zip(random_indices, random_seeds)
        )

        return p_value_adjust_func(p_values)
    else:
        return evaluate_kernel_test_on_samples(X, Y, Z, np.random.randint(np.iinfo(np.int32).max, size=1)[0])


def approx_kernel_based(
    X: np.ndarray,
    Y: np.ndarray,
    Z: Optional[np.ndarray] = None,
    num_random_features_X: int = 50,
    num_random_features_Y: int = 50,
    num_random_features_Z: int = 50,
    num_permutations: int = 100,
    approx_kernel: Callable[[np.ndarray], np.ndarray] = approximate_rbf_kernel_features,
    scale_data: bool = False,
    use_bootstrap: bool = True,
    bootstrap_num_runs: int = 10,
    bootstrap_num_samples: int = 1000,
    bootstrap_n_jobs: Optional[int] = None,
    p_value_adjust_func: Callable[[Union[np.ndarray, List[float]]], float] = quantile_based_fwer,
) -> float:
    """Implementation of the Randomized Conditional Independence Test. The independence test estimates a p-value
    for the null hypothesis that X and Y are independent (given Z). Depending whether Z is given, a conditional or
    pairwise independence test is performed.

    If Z is given: Using RCIT as conditional independence test.
    If Z is not given: Using RIT as pairwise independence test.

    Note:
    - The data can be multivariate, i.e. the given input matrices can have multiple columns.
    - Categorical data need to be represented as strings.
    - It is possible to apply a different kernel to each column in the matrices. For instance, a RBF kernel for the
      first dimension in X and a delta kernel for the second.

    Based on the work:
        Strobl, Eric V., Kun Zhang, and Shyam Visweswaran.
        Approximate kernel-based conditional independence tests for fast non-parametric causal discovery.
        Journal of Causal Inference 7.1 (2019).

    :param X: Data matrix for observations from X.
    :param Y: Data matrix for observations from Y.
    :param Z: Optional data matrix for observations from Z. This is the conditional variable.
    :param num_random_features_X: Number of features sampled from the approximated kernel map for X.
    :param num_random_features_Y: Number of features sampled from the approximated kernel map for Y.
    :param num_random_features_Z: Number of features sampled from the approximated kernel map for Z.
    :param num_permutations: Number of permutations for estimating the test test statistic.
    :param approx_kernel: The approximated kernel map. The expected input is a n x d numpy array and the output is
                          expected to be a n x k numpy array with k << d. By default, the Nystroem method with a RBF
                          kernel is used.
    :param scale_data: If set to True, the data will be standardized. If set to False, the data is taken as it is.
                       Standardizing the data helps in identifying weak dependencies. If one is only interested in
                       stronger ones, consider setting this to False.
    :param use_bootstrap: If True, the independence tests are performed on multiple subsets of the data and the final
                          p-value is constructed based on the provided p_value_adjust_func function.
    :param bootstrap_num_runs: Number of bootstrap runs (only relevant if use_bootstrap is True).
    :param bootstrap_num_samples: Maximum number of used samples per bootstrap run.
    :param bootstrap_n_jobs: Number of parallel jobs for the bootstrap runs.
    :param p_value_adjust_func: A callable that expects a numpy array of multiple p-values and returns one p-value. This
                                is typically used a family wise error rate control method.
    :return: The p-value for the null hypothesis that X and Y are independent (given Z).
    """
    bootstrap_n_jobs = config.default_n_jobs if bootstrap_n_jobs is None else bootstrap_n_jobs

    X = _remove_constant_columns(X)
    Y = _remove_constant_columns(Y)

    if X.shape[1] == 0 or Y.shape[1] == 0:
        # Either X and/or Y is constant.
        return 1.0

    if Z is not None:
        Z = _remove_constant_columns(Z)
        if Z.shape[1] == 0:
            # If Z is empty, we are in the pairwise setting.
            Z = None

    if not use_bootstrap:
        bootstrap_num_runs = 1
        bootstrap_num_samples = float("inf")
        bootstrap_n_jobs = 1

    if Z is None:
        return _rit(
            X,
            Y,
            num_permutations=num_permutations,
            num_random_features_X=num_random_features_X,
            num_random_features_Y=num_random_features_Y,
            num_runs=bootstrap_num_runs,
            num_max_samples_per_run=bootstrap_num_samples,
            approx_kernel=approx_kernel,
            scale_data=scale_data,
            n_jobs=bootstrap_n_jobs,
            p_value_adjust_func=p_value_adjust_func,
        )
    else:
        return _rcit(
            X,
            Y,
            Z,
            num_permutations=num_permutations,
            num_random_features_X=num_random_features_X,
            num_random_features_Y=num_random_features_Y,
            num_random_features_Z=num_random_features_Z,
            num_runs=bootstrap_num_runs,
            num_max_samples_per_run=bootstrap_num_samples,
            approx_kernel=approx_kernel,
            scale_data=scale_data,
            n_jobs=bootstrap_n_jobs,
            p_value_adjust_func=p_value_adjust_func,
        )


def _rit(
    X: np.ndarray,
    Y: np.ndarray,
    num_random_features_X: int,
    num_random_features_Y: int,
    num_permutations: int,
    num_runs: int,
    num_max_samples_per_run: int,
    approx_kernel: Callable[[np.ndarray, int], np.ndarray],
    scale_data: bool,
    n_jobs: Optional[int],
    p_value_adjust_func: Callable[[Union[np.ndarray, List[float]]], float],
) -> float:
    """Implementation of the Randomized Independence Test based on the work:
    Strobl, Eric V., Kun Zhang, and Shyam Visweswaran.
    Approximate kernel-based conditional independence tests for fast non-parametric causal discovery.
    Journal of Causal Inference 7.1 (2019).
    """
    n_jobs = config.default_n_jobs if n_jobs is None else n_jobs

    X, Y = _convert_to_numeric(*shape_into_2d(X, Y))

    if scale_data:
        X = scale(X)
        Y = scale(Y)

    def evaluate_rit_on_samples(parallel_random_seed: int):
        set_random_seed(parallel_random_seed)

        if X.shape[0] > num_max_samples_per_run:
            random_indices = np.random.choice(X.shape[0], num_max_samples_per_run, replace=False)
            X_samples = X[random_indices]
            Y_samples = Y[random_indices]
        else:
            X_samples = X
            Y_samples = Y

        random_features_x = scale(approx_kernel(X_samples, num_random_features_X))
        random_features_y = scale(approx_kernel(Y_samples, num_random_features_Y))

        permutation_results_of_statistic = []
        for i in range(num_permutations):
            permutation_results_of_statistic.append(
                _estimate_rit_statistic(
                    random_features_x[
                        np.random.choice(random_features_x.shape[0], random_features_x.shape[0], replace=False)
                    ],
                    random_features_y,
                )
            )

        return 1 - (
            np.sum(_estimate_rit_statistic(random_features_x, random_features_y) > permutation_results_of_statistic)
            / len(permutation_results_of_statistic)
        )

    random_seeds = np.random.randint(np.iinfo(np.int32).max, size=num_runs)
    p_values = Parallel(n_jobs=n_jobs)(delayed(evaluate_rit_on_samples)(random_seeds[i]) for i in range(num_runs))

    return p_value_adjust_func(p_values)


def _rcit(
    X: np.ndarray,
    Y: np.ndarray,
    Z: np.ndarray,
    num_random_features_X: int,
    num_random_features_Y: int,
    num_random_features_Z: int,
    num_permutations: int,
    num_runs: int,
    num_max_samples_per_run: int,
    approx_kernel: Callable[[np.ndarray, int], np.ndarray],
    scale_data: bool,
    n_jobs: Optional[int],
    p_value_adjust_func: Callable[[Union[np.ndarray, List[float]]], float],
) -> float:
    """
    Implementation of the Randomized Conditional Independence Test based on the work:
        Strobl, Eric V., Kun Zhang, and Shyam Visweswaran.
        Approximate kernel-based conditional independence tests for fast non-parametric causal discovery.
        Journal of Causal Inference 7.1 (2019).
    """
    n_jobs = config.default_n_jobs if n_jobs is None else n_jobs

    X, Y, Z = _convert_to_numeric(*shape_into_2d(X, Y, Z))

    if scale_data:
        X = scale(X)
        Y = scale(Y)
        Z = scale(Z)

    def parallel_job(parallel_random_seed: int):
        set_random_seed(parallel_random_seed)

        if X.shape[0] > num_max_samples_per_run:
            random_indices = np.random.choice(X.shape[0], num_max_samples_per_run, replace=False)
            X_samples = X[random_indices]
            Y_samples = Y[random_indices]
            Z_samples = Z[random_indices]
        else:
            X_samples = X
            Y_samples = Y
            Z_samples = Z

        Y_samples = np.column_stack([Y_samples, Z_samples])
        random_features_x = scale(approx_kernel(X_samples, num_random_features_X))
        random_features_y = scale(approx_kernel(Y_samples, num_random_features_Y))
        random_features_z = scale(approx_kernel(Z_samples, num_random_features_Z))

        cov_zz = _estimate_column_wise_covariances(random_features_z, random_features_z)
        inverse_cov_zz = scipy.linalg.cho_solve(
            scipy.linalg.cho_factor(cov_zz + np.eye(cov_zz.shape[0]) * 10**-10, lower=True), np.eye(cov_zz.shape[0])
        )
        cov_xz = _estimate_column_wise_covariances(random_features_x, random_features_z)
        cov_zy = _estimate_column_wise_covariances(random_features_z, random_features_y)

        z_inverse_cov_zz = random_features_z @ inverse_cov_zz

        residual_x = random_features_x - z_inverse_cov_zz @ cov_xz.T
        residual_y = random_features_y - z_inverse_cov_zz @ cov_zy

        # Estimate test statistic multiple times on different permutations of the data. The p-value is then the
        # probability (i.e. fraction) of obtaining a test statistic that is greater than statistic on the non-permuted
        # data.
        permutation_results_of_statistic = []
        for i in range(num_permutations):
            permutation_results_of_statistic.append(
                _estimate_rit_statistic(
                    residual_x[np.random.choice(residual_x.shape[0], residual_x.shape[0], replace=False)], residual_y
                )
            )

        return 1 - (
            np.sum(_estimate_rit_statistic(residual_x, residual_y) > permutation_results_of_statistic)
            / len(permutation_results_of_statistic)
        )

    random_seeds = np.random.randint(np.iinfo(np.int32).max, size=num_runs)
    p_values = Parallel(n_jobs=n_jobs)(delayed(parallel_job)(random_seeds[i]) for i in range(num_runs))

    return p_value_adjust_func(p_values)


def _estimate_rit_statistic(X: np.ndarray, Y: np.ndarray) -> np.ndarray:
    return X.shape[0] * np.sum(_estimate_column_wise_covariances(X, Y) ** 2)


def _estimate_column_wise_covariances(X: np.ndarray, Y: np.ndarray) -> np.ndarray:
    return np.cov(X, Y, rowvar=False)[: X.shape[1], -Y.shape[1] :]


def _convert_to_numeric(*args) -> List[np.ndarray]:
    result = []
    for X in args:
        X = np.array(X)
        for col in range(X.shape[1]):
            if isinstance(X[0, col], bool):
                X[:, col] = X[:, col].astype(str)

        result.append(apply_one_hot_encoding(X, fit_one_hot_encoders(X)))

    return result


def _remove_constant_columns(X: np.ndarray) -> np.ndarray:
    X = shape_into_2d(X)
    return X[:, [np.unique(X[:, i]).shape[0] > 1 for i in range(X.shape[1])]]



import networkx as nx
import numpy as np
import pandas as pd
from flaky import flaky
from pytest import approx
from scipy import stats

from dowhy.gcm import (
    AdditiveNoiseModel,
    ClassifierFCM,
    ScipyDistribution,
    StructuralCausalModel,
    auto,
    confidence_intervals,
    fit,
)
from dowhy.gcm.feature import feature_relevance_distribution, feature_relevance_sample, parent_relevance
from dowhy.gcm.ml import create_linear_regressor, create_logistic_regression_classifier
from dowhy.gcm.shapley import ShapleyApproximationMethods, ShapleyConfig
from dowhy.gcm.uncertainty import estimate_entropy_of_probabilities
from dowhy.gcm.util.general import means_difference


@flaky(max_runs=5)
def test_when_using_parent_relevance_with_continous_data_then_returns_correct_results():
    causal_model = StructuralCausalModel(nx.DiGraph([("X1", "X2"), ("X0", "X2")]))
    causal_model.set_causal_mechanism("X1", ScipyDistribution(stats.norm, loc=0, scale=1))
    causal_model.set_causal_mechanism("X0", ScipyDistribution(stats.norm, loc=0, scale=1))
    causal_model.set_causal_mechanism("X2", AdditiveNoiseModel(prediction_model=create_linear_regressor()))

    X0 = np.random.normal(0, 1, 1000)
    X1 = np.random.normal(0, 1, 1000)
    training_data = pd.DataFrame({"X0": X0, "X1": X1, "X2": 3 * X0 + X1})

    fit(causal_model, training_data)
    relevance, noise = parent_relevance(causal_model, "X2")

    # Contributions should add up to Var(X2)
    assert relevance[("X0", "X2")] + relevance[("X1", "X2")] + noise == approx(
        np.var(training_data["X2"].to_numpy()), abs=1.5
    )
    assert relevance[("X0", "X2")] == approx(9, abs=1)
    assert relevance[("X1", "X2")] == approx(1, abs=0.3)
    assert noise == approx(0, abs=0.5)


@flaky(max_runs=5)
def test_when_using_parent_relevance_with_categorical_data_then_returns_correct_results():
    causal_model = StructuralCausalModel(nx.DiGraph([("X0", "Y"), ("X1", "Y"), ("X2", "Y"), ("X3", "Y"), ("X4", "Y")]))
    causal_model.set_causal_mechanism("X0", ScipyDistribution(stats.uniform, loc=0, scale=1))
    causal_model.set_causal_mechanism("X1", ScipyDistribution(stats.uniform, loc=0, scale=1))
    causal_model.set_causal_mechanism("X2", ScipyDistribution(stats.uniform, loc=0, scale=1))
    causal_model.set_causal_mechanism("X3", ScipyDistribution(stats.uniform, loc=0, scale=1))
    causal_model.set_causal_mechanism("X4", ScipyDistribution(stats.uniform, loc=0, scale=1))
    causal_model.set_causal_mechanism("Y", ClassifierFCM())

    X = np.random.uniform(0, 1, (1000, 5))
    Y = []

    for n in X:
        if n[0] + n[1] > 1:
            Y.append(0)
        else:
            Y.append(1)

    X0 = X[:, 0]
    X1 = X[:, 1]
    X2 = X[:, 2]
    X3 = X[:, 3]
    X4 = X[:, 4]
    Y = np.array(Y).astype(str)

    fit(causal_model, data=pd.DataFrame({"X0": X0, "X1": X1, "X2": X2, "X3": X3, "X4": X4, "Y": Y}))

    relevance, noise = parent_relevance(causal_model, "Y", num_samples_randomization=1000, num_samples_baseline=100)

    assert relevance[("X0", "Y")] == approx(0.125, abs=0.05)
    assert relevance[("X1", "Y")] == approx(0.125, abs=0.05)
    assert relevance[("X2", "Y")] == approx(0, abs=0.05)
    assert relevance[("X3", "Y")] == approx(0, abs=0.05)
    assert relevance[("X4", "Y")] == approx(0, abs=0.05)
    assert noise == approx(0, abs=0.05)


@flaky(max_runs=5)
def test_when_using_parent_relevance_with_confidence_intervals_then_returns_reasonable_bounds():
    causal_model = StructuralCausalModel(nx.DiGraph([("X1", "X2"), ("X0", "X2")]))
    causal_model.set_causal_mechanism("X1", ScipyDistribution(stats.norm, loc=0, scale=1))
    causal_model.set_causal_mechanism("X0", ScipyDistribution(stats.norm, loc=0, scale=1))
    causal_model.set_causal_mechanism("X2", AdditiveNoiseModel(prediction_model=create_linear_regressor()))

    X0 = np.random.normal(0, 1, 1000)
    X1 = np.random.normal(0, 1, 1000)

    training_data = pd.DataFrame({"X0": X0, "X1": X1, "X2": 3 * X0 + X1})
    fit(causal_model, training_data)

    def estimation_func():
        dict_result, noise = parent_relevance(causal_model, "X2")
        dict_result[("noise", "X2")] = noise
        return dict_result

    median_relevance, cis = confidence_intervals(estimation_func, num_bootstrap_resamples=10)

    # Contributions should add up to Var(X2)
    assert median_relevance[("X0", "X2")] == approx(9, abs=1)
    assert median_relevance[("X1", "X2")] == approx(1, abs=0.3)
    assert median_relevance[("noise", "X2")] == approx(0, abs=0.5)

    assert cis[("X0", "X2")] == approx(np.array([8.5, 9.5]), abs=1)
    assert cis[("X1", "X2")] == approx(np.array([0.8, 1.2]), abs=0.4)
    assert cis[("noise", "X2")] == approx(np.array([-0.2, 0.2]), abs=0.4)


@flaky(max_runs=5)
def test_feature_relevance_sample_mean_diff():
    num_vars = 15
    X = np.random.normal(0, 1, (1000, num_vars))
    coefficients = np.random.choice(20, num_vars) - 10

    def model(x) -> np.ndarray:
        return np.sum(coefficients * x, axis=1)

    shapley_values = feature_relevance_sample(
        model, feature_samples=X, baseline_samples=X[:20], subset_scoring_func=means_difference
    )

    for i in range(20):
        assert coefficients * (X[i, :] - np.mean(X, axis=0)) == approx(shapley_values[i], abs=0.001)


@flaky(max_runs=3)
def test_given_baseline_values_when_estimating_feature_relevance_sample_with_mean_diff_then_returns_expected_result():
    num_vars = 5
    X = np.random.normal(0, 1, (1000, num_vars))
    coefficients = np.random.choice(20, num_vars) - 10

    def model(x) -> np.ndarray:
        return np.sum(coefficients * x, axis=1)

    shapley_values_1 = feature_relevance_sample(
        model,
        feature_samples=X,
        baseline_samples=X[:20],
        subset_scoring_func=lambda x, y: np.mean(x) * y,
        baseline_target_values=np.zeros(20),
        shapley_config=ShapleyConfig(approximation_method=ShapleyApproximationMethods.EXACT),
    )

    shapley_values_2 = feature_relevance_sample(
        model,
        feature_samples=X,
        baseline_samples=X[:20],
        subset_scoring_func=lambda x, y: np.mean(x) * y,
        baseline_target_values=np.zeros(20) + 1,
        shapley_config=ShapleyConfig(approximation_method=ShapleyApproximationMethods.EXACT),
    )

    shapley_values_3 = feature_relevance_sample(
        model,
        feature_samples=X,
        baseline_samples=X[:20],
        subset_scoring_func=lambda x, y: np.mean(x) * y,
        baseline_target_values=np.zeros(20) + 2,
        shapley_config=ShapleyConfig(approximation_method=ShapleyApproximationMethods.EXACT),
    )

    for i in range(20):
        assert shapley_values_1[i] == approx(np.zeros(5))
        assert shapley_values_2[i] * 2 == approx(shapley_values_3[i])


@flaky(max_runs=5)
def test_feature_relevance_sample_mean_diff_with_certain_batch_size():
    X = np.random.normal(0, 1, (1000, 3))
    coefficients = np.random.choice(20, 3) - 10

    def model(x) -> np.ndarray:
        return np.sum(coefficients * x, axis=1)

    shapley_values = feature_relevance_sample(
        model, feature_samples=X, baseline_samples=X, subset_scoring_func=means_difference, max_batch_size=123
    )

    for i in range(1000):
        assert coefficients * (X[i, :] - np.mean(X, axis=0)) == approx(shapley_values[i], abs=0.001)


@flaky(max_runs=3)
def test_when_using_feature_relevance_distribution_with_entropy_set_function_then_returns_correct_results():
    X = np.random.uniform(0, 1, (3000, 5))
    Y = []

    for n in X:
        if n[0] + n[1] > 1:
            Y.append(0)
        else:
            Y.append(1)

    Y = np.array(Y).astype(str)

    classifier_mdl = create_logistic_regression_classifier()
    classifier_mdl.fit(X, Y)

    # H(P(Y)) -- Can be precomputed
    h_p_Y = estimate_entropy_of_probabilities(np.mean(classifier_mdl.predict_probabilities(X), axis=0).reshape(1, -1))

    # -(H(P(Y | do(x_S)) - H(P(Y))) = H(P(Y)) - H(P(Y | do(x_S))
    def my_entropy_feature_attribution_function(
        randomized_predictions: np.ndarray, baseline_predictions: np.ndarray
    ) -> float:
        # H(P(Y | do(x_S)) = H(E[P(Y | x_S, X'_\S)])

        # E[P(Y | x_S, X'_\S)]
        p_y_do_xs = np.mean(randomized_predictions, axis=0).reshape(1, -1)

        # H(E[P(Y | x_S, X'_\S)])
        h_p_Y_do_xs = estimate_entropy_of_probabilities(p_y_do_xs)

        # Using H(P(Y)) based on the origina data, i.e. ignoring baseline_predictions.
        return h_p_Y - h_p_Y_do_xs

    feature_contributions = feature_relevance_distribution(
        classifier_mdl.predict_probabilities, X, my_entropy_feature_attribution_function
    )

    # E[H(P(Y)) - H(P(Y | do(X_U))] = H(P(Y)) - E[H(P(Y | X))]
    expected_sum_shapley_values = h_p_Y - estimate_entropy_of_probabilities(classifier_mdl.predict_probabilities(X))

    assert feature_contributions[0] == approx(expected_sum_shapley_values / 2, abs=0.075)
    assert feature_contributions[1] == approx(expected_sum_shapley_values / 2, abs=0.075)
    assert feature_contributions[2] == approx(0, abs=0.01)
    assert feature_contributions[3] == approx(0, abs=0.01)
    assert feature_contributions[4] == approx(0, abs=0.01)

    assert np.sum(feature_contributions) == approx(expected_sum_shapley_values, abs=0.04)


@flaky(max_runs=3)
def test_given_misspecified_graph_when_estimating_parent_relevance_with_observed_data_then_returns_correct_result():
    causal_model_without = StructuralCausalModel(nx.DiGraph([("X0", "X2"), ("X1", "X2")]))
    causal_model_with = StructuralCausalModel(nx.DiGraph([("X0", "X1"), ("X0", "X2"), ("X1", "X2")]))

    X0 = np.random.normal(0, 1, 1000)
    X1 = X0 + np.random.normal(0, 1, 1000)

    training_data = pd.DataFrame({"X0": X0, "X1": X1, "X2": X0 + X1})
    auto.assign_causal_mechanisms(causal_model_without, training_data, auto.AssignmentQuality.GOOD)
    auto.assign_causal_mechanisms(causal_model_with, training_data, auto.AssignmentQuality.GOOD)

    fit(causal_model_without, training_data)
    fit(causal_model_with, training_data)

    feature_relevance_missing_edge, _ = parent_relevance(causal_model_without, "X2", training_data)
    feature_relevance_with_edge, _ = parent_relevance(causal_model_with, "X2")

    assert feature_relevance_missing_edge[("X0", "X2")] == approx(feature_relevance_with_edge[("X0", "X2")], abs=0.15)
    assert feature_relevance_missing_edge[("X1", "X2")] == approx(feature_relevance_with_edge[("X1", "X2")], abs=0.15)



import networkx as nx
import numpy as np
import pandas as pd
import pytest
from flaky import flaky
from pytest import approx

from dowhy.gcm import (
    AdditiveNoiseModel,
    EmpiricalDistribution,
    ProbabilisticCausalModel,
    draw_samples,
    fit,
    is_root_node,
)
from dowhy.gcm.ml import create_linear_regressor


@flaky(max_runs=2)
def test_fit_and_draw_samples():
    X0 = np.random.uniform(-1, 1, 1000)
    X1 = 2 * X0 + np.random.normal(0, 0.1, 1000)
    X2 = 0.5 * X0 + np.random.normal(0, 0.1, 1000)
    X3 = 0.5 * X2 + np.random.normal(0, 0.1, 1000)

    original_observations = pd.DataFrame({"X0": X0, "X1": X1, "X2": X2, "X3": X3})

    scm = ProbabilisticCausalModel(nx.DiGraph([("X0", "X1"), ("X0", "X2"), ("X2", "X3")]))
    scm.set_causal_mechanism("X0", EmpiricalDistribution())
    scm.set_causal_mechanism("X1", AdditiveNoiseModel(prediction_model=create_linear_regressor()))
    scm.set_causal_mechanism("X2", AdditiveNoiseModel(prediction_model=create_linear_regressor()))
    scm.set_causal_mechanism("X3", AdditiveNoiseModel(prediction_model=create_linear_regressor()))

    fit(scm, original_observations)

    generated_samples = draw_samples(scm, 1000)

    assert np.mean(generated_samples["X0"]) == approx(np.mean(X0), abs=0.1)
    assert np.std(generated_samples["X0"]) == approx(np.std(X0), abs=0.2)
    assert np.mean(generated_samples["X1"]) == approx(np.mean(X1), abs=0.1)
    assert np.std(generated_samples["X1"]) == approx(np.std(X1), abs=0.2)
    assert np.mean(generated_samples["X2"]) == approx(np.mean(X2), abs=0.1)
    assert np.std(generated_samples["X2"]) == approx(np.std(X2), abs=0.2)
    assert np.mean(generated_samples["X3"]) == approx(np.mean(X3), abs=0.1)
    assert np.std(generated_samples["X3"]) == approx(np.std(X3), abs=0.2)


def test_set_causal_model_raises_error():
    with pytest.raises(ValueError):
        ProbabilisticCausalModel().set_causal_mechanism("X0", EmpiricalDistribution())


def test_is_root_node():
    graph = nx.DiGraph([("X", "Z"), ("Y", "Z")])
    assert is_root_node(graph, "X") == True
    assert is_root_node(graph, "Y") == True
    assert is_root_node(graph, "Z") == False



import pytest

from dowhy.causal_graph import CausalGraph
from dowhy.causal_identifier import AutoIdentifier, BackdoorAdjustment
from dowhy.causal_identifier.identify_effect import EstimandType

from .base import IdentificationTestGraphSolution, example_graph_solution


class TestBackdoorIdentification(object):
    def test_identify_backdoor_no_biased_sets(self, example_graph_solution: IdentificationTestGraphSolution):
        graph = example_graph_solution.graph
        biased_sets = example_graph_solution.biased_sets
        identifier = AutoIdentifier(
            estimand_type=EstimandType.NONPARAMETRIC_ATE,
            backdoor_adjustment=BackdoorAdjustment.BACKDOOR_EXHAUSTIVE,
        )

        backdoor_results = identifier.identify_backdoor(graph, "X", "Y", include_unobserved=False)
        backdoor_sets = [
            set(backdoor_result_dict["backdoor_set"])
            for backdoor_result_dict in backdoor_results
            if len(backdoor_result_dict["backdoor_set"]) > 0
        ]

        assert (len(backdoor_sets) == 0 and len(biased_sets) == 0) or all(  # No biased sets exist and that's expected.
            [set(biased_backdoor_set) not in backdoor_sets for biased_backdoor_set in biased_sets]
        )  # No sets that would induce biased results are present in the solution.

    def test_identify_backdoor_unobserved_not_in_backdoor_set(
        self, example_graph_solution: IdentificationTestGraphSolution
    ):
        graph = example_graph_solution.graph
        observed_variables = example_graph_solution.observed_variables
        identifier = AutoIdentifier(
            estimand_type=EstimandType.NONPARAMETRIC_ATE,
            backdoor_adjustment=BackdoorAdjustment.BACKDOOR_EXHAUSTIVE,
        )

        backdoor_results = identifier.identify_backdoor(graph, "X", "Y", include_unobserved=False)
        backdoor_sets = [
            set(backdoor_result_dict["backdoor_set"])
            for backdoor_result_dict in backdoor_results
            if len(backdoor_result_dict["backdoor_set"]) > 0
        ]

        assert all(
            [variable in observed_variables for backdoor_set in backdoor_sets for variable in backdoor_set]
        )  # All variables used in the backdoor sets must be observed.

    def test_identify_backdoor_minimal_adjustment(self, example_graph_solution: IdentificationTestGraphSolution):
        graph = example_graph_solution.graph
        expected_sets = example_graph_solution.minimal_adjustment_sets
        identifier = AutoIdentifier(
            estimand_type=EstimandType.NONPARAMETRIC_ATE,
            backdoor_adjustment=BackdoorAdjustment.BACKDOOR_MIN,
            proceed_when_unidentifiable=False,
        )

        backdoor_results = identifier.identify_backdoor(graph, "X", "Y", include_unobserved=False)
        backdoor_sets = [set(backdoor_result_dict["backdoor_set"]) for backdoor_result_dict in backdoor_results]

        assert (
            (len(backdoor_sets) == 0) and (len(expected_sets) == 0)
        ) or all(  # No adjustments exist and that's expected.
            [set(expected_set) in backdoor_sets for expected_set in expected_sets]
        )

    def test_identify_backdoor_maximal_adjustment(self, example_graph_solution: IdentificationTestGraphSolution):
        graph = example_graph_solution.graph
        expected_sets = example_graph_solution.maximal_adjustment_sets
        identifier = AutoIdentifier(
            estimand_type=EstimandType.NONPARAMETRIC_ATE,
            backdoor_adjustment=BackdoorAdjustment.BACKDOOR_MAX,
            proceed_when_unidentifiable=False,
        )

        backdoor_results = identifier.identify_backdoor(graph, "X", "Y", include_unobserved=False)

        backdoor_sets = [set(backdoor_result_dict["backdoor_set"]) for backdoor_result_dict in backdoor_results]
        print(backdoor_sets, expected_sets, example_graph_solution.graph_str)
        assert (
            (len(backdoor_sets) == 0) and (len(expected_sets) == 0)
        ) or all(  # No adjustments exist and that's expected.
            [set(expected_set) in backdoor_sets for expected_set in expected_sets]
        )

    def test_identify_backdoor_maximal_direct_effect(self, example_graph_solution: IdentificationTestGraphSolution):
        graph = example_graph_solution.graph
        expected_sets = example_graph_solution.direct_maximal_adjustment_sets
        identifier = AutoIdentifier(
            estimand_type=EstimandType.NONPARAMETRIC_CDE,
            backdoor_adjustment=BackdoorAdjustment.BACKDOOR_MAX,
            proceed_when_unidentifiable=False,
        )

        backdoor_results = identifier.identify_backdoor(graph, "X", "Y", direct_effect=True)

        backdoor_sets = [set(backdoor_result_dict["backdoor_set"]) for backdoor_result_dict in backdoor_results]
        print(backdoor_sets, expected_sets, example_graph_solution.graph_str)
        assert (
            (len(backdoor_sets) == 0) and (len(expected_sets) == 0)
        ) or all(  # No adjustments exist and that's expected.
            [set(expected_set) in backdoor_sets for expected_set in expected_sets]
        )



from typing import Any, Callable, List, Optional, Union

import pandas as pd

from dowhy.causal_estimator import CausalEstimate
from dowhy.causal_identifier.identified_estimand import IdentifiedEstimand
from dowhy.causal_refuter import CausalRefutation
from dowhy.causal_refuters.add_unobserved_common_cause import sensitivity_simulation
from dowhy.causal_refuters.bootstrap_refuter import refute_bootstrap
from dowhy.causal_refuters.data_subset_refuter import refute_data_subset
from dowhy.causal_refuters.dummy_outcome_refuter import refute_dummy_outcome
from dowhy.causal_refuters.placebo_treatment_refuter import refute_placebo_treatment
from dowhy.causal_refuters.random_common_cause import refute_random_common_cause

ALL_REFUTERS = [
    sensitivity_simulation,
    refute_bootstrap,
    refute_data_subset,
    refute_dummy_outcome,
    refute_placebo_treatment,
    refute_random_common_cause,
    # TODO: Sensitivity Analyzers excluded from list due to different return type
]


def refute_estimate(
    data: pd.DataFrame,
    target_estimand: IdentifiedEstimand,
    estimate: CausalEstimate,
    treatment_name: Optional[str] = None,
    outcome_name: Optional[str] = None,
    refuters: List[Callable[..., Union[CausalRefutation, List[CausalRefutation]]]] = ALL_REFUTERS,
    **kwargs,
) -> List[CausalRefutation]:
    """Executes a list of refuters using the default parameters
       Only refuters that return CausalRefutation or a list of CausalRefutation is supported

    :param data: pd.DataFrame: Data to run the refutation
    :param target_estimand: IdentifiedEstimand: Identified estimand to run the refutation
    :param estimate: CausalEstimate: Estimate to run the refutation
    :param treatment_name: str: Name of the treatment (Optional)
    :param outcome_name: str: Name of the outcome (Optional)
    :param refuters: list: List of refuters to execute
    :**kwargs: Replace any default for the provided list of refuters

    """
    refuter_kwargs = {
        "data": data,
        "target_estimand": target_estimand,
        "estimate": estimate,
        "treatment_name": treatment_name,
        "outcome_name": outcome_name,
        **kwargs,
    }

    results = []
    for refuter in refuters:
        refute = refuter(**refuter_kwargs)
        if isinstance(refute, list):
            results.extend(refute)
        else:
            results.append(refute)

    return results



import networkx as nx

from dowhy.utils.graph_operations import adjacency_matrix_to_adjacency_list


class NodePair:
    """
    Data structure to store backdoor variables between 2 nodes.
    """

    def __init__(self, node1, node2):
        self._node1 = node1
        self._node2 = node2
        self._is_blocked = None  # To store if all paths between node1 and node2 are blocked
        self._condition_vars = []  # To store variable to be conditioned on to block all paths between node1 and node2
        self._complete = False  # To store to paths between node pair have been completely explored.

    def update(self, path, condition_vars=None):
        if condition_vars is None:
            """path is a Path variable"""
            if self._is_blocked is None:
                self._is_blocked = path.is_blocked()
            else:
                self._is_blocked = self._is_blocked and path.is_blocked()
            if not path.is_blocked():
                self._condition_vars.append(path.get_condition_vars())

        else:
            """path is a list"""
            condition_vars = list(condition_vars)
            self._condition_vars.append(set([*path[1:], *condition_vars]))

    def get_condition_vars(self):
        return self._condition_vars

    def set_complete(self):
        self._complete = True

    def is_complete(self):
        return self._complete

    def __str__(self):
        string = ""
        string += "Blocked: " + str(self._is_blocked) + "\n"
        if not self._is_blocked:
            condition_vars = [str(s) for s in self._condition_vars]
            string += "To block path, condition on: " + ",".join(condition_vars) + "\n"
        return string


class Path:
    """
    Data structure to store a particular path between 2 nodes.
    """

    def __init__(self):
        self._is_blocked = None  # To store if path is blocked
        self._condition_vars = set()  # To store variables needed to block the path

    def update(self, path, is_blocked):
        """
        path is a list
        """
        self._is_blocked = is_blocked
        if not is_blocked:
            self._condition_vars = self._condition_vars.union(set(path[1:-1]))

    def is_blocked(self):
        return self._is_blocked

    def get_condition_vars(self):
        return self._condition_vars

    def __str__(self):
        string = ""
        string += "Blocked: " + str(self._is_blocked) + "\n"
        if not self._is_blocked:
            string += "To block path, condition on: " + ",".join(self._condition_vars) + "\n"
        return string


class Backdoor:
    """
    Class for optimized implementation of Backdoor variable search between the source nodes and the target nodes.
    """

    def __init__(self, graph, nodes1, nodes2):
        self._graph = graph
        self._nodes1 = nodes1
        self._nodes2 = nodes2
        self._nodes12 = set(self._nodes1).union(self._nodes2)  # Total set of nodes
        self._colliders = set()

    def get_backdoor_vars(self):
        """
        Obtains sets of backdoor variable to condition on for each node pair.

        :returns:  List of sets with each set containing backdoor variable corresponding to a given node pair.
        """
        undirected_graph = self._graph.to_undirected()

        # Get adjacency list
        adjlist = adjacency_matrix_to_adjacency_list(
            nx.to_numpy_matrix(undirected_graph), labels=list(undirected_graph.nodes)
        )
        path_dict = {}
        backdoor_sets = []  # Put in backdoor sets format

        for node1 in self._nodes1:
            for node2 in self._nodes2:
                if (node1, node2) in path_dict:
                    continue
                self._path_search(adjlist, node1, node2, path_dict)
                if len(path_dict) != 0:
                    obj = HittingSetAlgorithm(path_dict[(node1, node2)].get_condition_vars(), self._colliders)

                    backdoor_set = {}
                    backdoor_set["backdoor_set"] = tuple(obj.find_set())
                    backdoor_set["num_paths_blocked_by_observed_nodes"] = obj.num_sets()
                    backdoor_sets.append(backdoor_set)

        return backdoor_sets

    def is_backdoor(self, path):
        """
        Check if path is a backdoor path.

        :param path: List of nodes comprising the path.
        """
        if len(path) < 2:
            return False
        return True if self._graph.has_edge(path[1], path[0]) else False

    def _path_search_util(self, graph, node1, node2, vis, path, path_dict, is_blocked=False, prev_arrow=None):
        """
        :param graph: Adjacency list of the graph under consideration.
        :param node1: Current node being considered.
        :param node2: Target node.
        :param vis: Set of already visited nodes.
        :param path: List of nodes comprising the path upto node1.
        :path path_dict: Dictionary of node pairs.
        :param is_blocked: True is path is blocked by a collider, else False.
        :param prev_arrow: Described state of previous arrow. True if arrow incoming, False if arrow outgoing.
        """
        if is_blocked:
            return

        # If node pair has been fully explored
        if ((node1, node2) in path_dict) and (path_dict[(node1, node2)].is_complete()):
            for i in range(len(path)):
                if (path[i], node2) not in path_dict:
                    path_dict[(path[i], node2)] = NodePair(path[i], node2)
                obj = HittingSetAlgorithm(path_dict[(node1, node2)].get_condition_vars(), self._colliders)
                # Add node1 to backdoor set of node_pair
                s = set([node1])
                s = s.union(obj.find_set())
                path_dict[(path[i], node2)].update(path[i:], s)

        else:
            path.append(node1)
            vis.add(node1)
            if node1 == node2:
                # Check if path is backdoor and does not have nodes1\node1 or nodes2\node2 as intermediate nodes
                if self.is_backdoor(path) and len(self._nodes12.intersection(set(path[1:-1]))) == 0:
                    for i in range(len(path) - 1):
                        if (path[i], node2) not in path_dict:
                            path_dict[(path[i], node2)] = NodePair(path[i], node2)
                        path_var = Path()
                        path_var.update(path[i:].copy(), is_blocked)
                        path_dict[(path[i], node2)].update(path_var)
            else:
                for neighbour in graph[node1]:
                    if neighbour not in vis:
                        # True if arrow incoming, False if arrow outgoing
                        next_arrow = False if self._graph.has_edge(node1, neighbour) else True
                        if next_arrow == True and prev_arrow == True:
                            is_blocked = True
                            self._colliders.add(node1)
                        self._path_search_util(
                            graph, neighbour, node2, vis, path, path_dict, is_blocked, not next_arrow
                        )  # Since incoming for current node is outgoing for the next
            path.pop()
            vis.remove(node1)

        # Mark pair (node1, node2) complete
        if (node1, node2) in path_dict:
            path_dict[(node1, node2)].set_complete()

    def _path_search(self, graph, node1, node2, path_dict):
        """
        Path search using DFS.

        :param graph: Adjacency list of the graph under consideration.
        :param node1: Current node being considered.
        :param node2: Target node.
        :path path_dict: Dictionary of node pairs.
        """
        vis = set()
        self._path_search_util(graph, node1, node2, vis, [], path_dict, is_blocked=False)


class HittingSetAlgorithm:
    """
    Class for the Hitting Set Algorithm to obtain a approximate minimal set of backdoor variables
    to condition on for each node pair.
    """

    def __init__(self, list_of_sets, colliders=set()):
        """
        :param list_of_sets: List of sets such that each set comprises nodes representing a single backdoor path between a source node and a target node.
        """
        self._list_of_sets = list_of_sets
        self._colliders = colliders
        self._var_count = self._count_vars()

    def num_sets(self):
        """
        Obtain number of backdoor paths between a node pair.
        """
        return len(self._list_of_sets)

    def find_set(self):
        """
        Find approximate minimal set of nodes such that there is atleast one node from each set in list_of_sets.

        :returns: Approximate minimal set of nodes.
        """
        var_set = set()
        num_indices = len(self._list_of_sets)
        indices_covered = set()
        all_set_indices = set([i for i in range(num_indices)])

        while not self._is_covered(indices_covered, num_indices):
            set_index = all_set_indices - indices_covered
            max_el = self._max_occurence_var(var_dict=self._var_count)
            if max_el is None:
                break
            var_set.add(max_el)

            # Modify variable count and indices covered
            covered_present = self._indices_covered(el=max_el, set_index=set_index)
            self._modify_count(covered_present)
            indices_covered = indices_covered.union(covered_present)
        return var_set

    def _count_vars(self, set_index=None):
        """
        Obtain count of number of sets each particular node belongs to.

        :param set_index: Set of indices to consider for calculating the number of sets "hit" by a variable..
        """
        var_dict = {}

        if set_index == None:
            set_index = set([i for i in range(len(self._list_of_sets))])

        for idx in set_index:
            s = self._list_of_sets[idx]

            for el in s:
                if el not in self._colliders:
                    if el not in var_dict:
                        var_dict[el] = 0
                    var_dict[el] += 1

        return var_dict

    def _modify_count(self, indices_covered):
        """
        Modify count of number of sets each particular node belongs to based on nodes already covered in the previous iteration of the algorithm.
        """
        for idx in indices_covered:
            for el in self._list_of_sets[idx]:
                if el not in self._colliders:
                    self._var_count[el] -= 1

    def _max_occurence_var(self, var_dict):
        """
        Find the node contained in most number of sets.
        """
        max_el = None
        max_count = 0
        for key, val in var_dict.items():
            if val > max_count:
                max_count = val
                max_el = key
        return max_el

    def _indices_covered(self, el, set_index=None):
        """
        Obtain indices covered in a particular iteration of the algorithm.
        """
        covered = set()
        if set_index == None:
            set_index = set([i for i in range(len(self._list_of_sets))])

        for idx in set_index:
            if el in self._list_of_sets[idx]:
                covered.add(idx)
        return covered

    def _is_covered(self, indices_covered, num_indices):
        """
        List of sets is covered by the variable set.
        """

        covered = [False for i in range(num_indices)]
        for idx in indices_covered:
            covered[idx] = True
        return all(covered)



import string
from importlib import import_module

from dowhy.graph_learner import GraphLearner


def get_discovery_class_object(method_name, *args, **kwargs):
    """
    Import class from graph_learners.

    """
    # from https://www.bnmetrics.com/blog/factory-pattern-in-python3-simple-version
    try:
        module_name = method_name
        class_name = module_name.upper()

        discovery_module = import_module("." + module_name, package="dowhy.graph_learners")
        discovery_class = getattr(discovery_module, class_name)
        assert issubclass(discovery_class, GraphLearner)

    except (AttributeError, AssertionError, ImportError):
        raise ImportError("{} is not an existing causal discovery method.".format(method_name))
    return discovery_class


def get_library_class_object(module_method_name, *args, **kwargs):
    """
    Import library for causal inference.

    """
    # from https://www.bnmetrics.com/blog/factory-pattern-in-python3-simple-version
    try:
        (module_name, _, class_name) = module_method_name.rpartition(".")
        discovery_module = import_module(module_name)
        discovery_class = getattr(discovery_module, class_name)

    except (AttributeError, AssertionError, ImportError):
        raise ImportError(
            "Error loading {}.{}. Double-check the method name and ensure that all library dependencies are installed.".format(
                module_name, class_name
            )
        )
    return discovery_class



from datetime import datetime

import matplotlib.pyplot as plt

SMALL_SIZE = 8
MEDIUM_SIZE = 26
BIGGER_SIZE = 30


def plot_treatment_outcome(treatment, outcome, time_var):
    plt.rc("font", size=SMALL_SIZE)  # controls default text sizes
    plt.rc("axes", titlesize=BIGGER_SIZE)  # fontsize of the axes title
    plt.rc("axes", labelsize=MEDIUM_SIZE)  # fontsize of the x and y labels
    plt.rc("xtick", labelsize=MEDIUM_SIZE)  # fontsize of the tick labels
    plt.rc("ytick", labelsize=MEDIUM_SIZE)  # fontsize of the tick labels
    plt.rc("legend", fontsize=MEDIUM_SIZE)  # legend fontsize
    plt.rc("figure", titlesize=BIGGER_SIZE)  # fontsize of the figure title
    fig, ax = plt.subplots()
    tline = ax.plot(time_var, treatment, "o", label="Treatment")
    oline = ax.plot(time_var, outcome, "r^", label="Outcome")

    ax.legend(loc="upper left", bbox_to_anchor=(1.04, 1))
    plt.xlabel("Time")
    fig.set_size_inches(8, 6)
    fig.savefig("obs_data" + datetime.now().strftime("%H-%M-%S") + ".png", bbox_inches="tight")


def plot_causal_effect(estimate, treatment, outcome):
    plt.rc("font", size=SMALL_SIZE)  # controls default text sizes
    plt.rc("axes", titlesize=BIGGER_SIZE)  # fontsize of the axes title
    plt.rc("axes", labelsize=MEDIUM_SIZE)  # fontsize of the x and y labels
    plt.rc("xtick", labelsize=MEDIUM_SIZE)  # fontsize of the tick labels
    plt.rc("ytick", labelsize=MEDIUM_SIZE)  # fontsize of the tick labels
    plt.rc("legend", fontsize=MEDIUM_SIZE)  # legend fontsize
    plt.rc("figure", titlesize=BIGGER_SIZE)  # fontsize of the figure title
    fig, ax = plt.subplots()
    x_min = 0
    x_max = max(treatment)
    y_min = estimate.params["intercept"]
    y_max = y_min + estimate.value * (x_max - x_min)
    ax.scatter(treatment, outcome, c="gray", marker="o", label="Observed data")
    ax.plot([x_min, x_max], [y_min, y_max], c="black", ls="solid", lw=4, label="Causal variation")
    ax.set_ylim(0, max(outcome))
    ax.set_xlim(0, x_max)

    bbox_props = dict(boxstyle="round", fc="w", ec="0.5", alpha=0.9)
    ax.text(
        10.8,
        1,
        r"DoWhy estimate $\rho$ (slope) = " + str(round(estimate.value, 2)),
        ha="right",
        va="bottom",
        size=20,
        bbox=bbox_props,
    )
    ax.legend(loc="upper left")
    plt.xlabel("Treatment")
    plt.ylabel("Outcome")

    fig.set_size_inches(8, 6)
    fig.savefig("effect" + datetime.now().strftime("%H-%M-%S") + ".png", bbox_inches="tight")
